from __future__ import annotations

import os
from typing import Optional, Union
from starlette.datastructures import UploadFile

from pyonir.core import PyonirRequest
from pyonir.models.database import BaseFSQuery
from pyonir.models.page import BaseFile

from enum import Enum, unique

@unique
class AudioFormat(Enum):
    MP3  = "mp3"
    WAV  = "wav"
    FLAC = "flac"
    AAC  = "aac"
    OGG  = "ogg"
    M4A  = "m4a"

    def __str__(self):
        return self.value

@unique
class VideoFormat(Enum):
    MP4  = "mp4"
    MKV  = "mkv"
    AVI  = "avi"
    MOV  = "mov"
    WEBM = "webm"
    FLV  = "flv"

    def __str__(self):
        return self.value

@unique
class ImageFormat(Enum):
    JPG   = "jpg"
    JPEG  = "jpeg"
    PNG   = "png"
    GIF   = "gif"
    BMP   = "bmp"
    TIFF  = "tiff"
    WEBP  = "webp"
    SVG   = "svg"

    def __str__(self):
        return self.value

class BaseMedia(BaseFile):
    """Represents an image file and its details."""

    def __init__(self, path: str = None):
        super().__init__(path)
        filename_meta = self.decode_filename(self.file_name)
        self.data = filename_meta or self.get_media_data(self.file_path)
        self.has_encoded_filename = bool(filename_meta)
        pass

    @property
    def file_name(self):
        name, ext = os.path.splitext(os.path.basename(self.file_path))
        self.file_ext = ext
        return name

    @property
    def file_dirpath(self):
        return os.path.dirname(self.file_path)

    @property
    def file_dirname(self):
        return os.path.basename(os.path.dirname(self.file_path))

    @property
    def slug(self):
        return f"{self.file_dirname}/{self.file_name}{self.file_ext}"

    def rename_media_file(self) -> None:
        """Renames media file as b64 encoded value"""
        encoded_filename = self.encode_filename(self.file_path, self.data)
        new_filepath = self.file_path.replace(self.file_name+self.file_ext, encoded_filename)
        os.rename(self.file_path, new_filepath)
        self.file_path = new_filepath

    def open_image(self):
        from PIL import Image
        raw_img = Image.open(self.file_path)
        return raw_img

    def resize(self, sizes=None):
        '''
        Resize each image and save to the upload path in corresponding image size and paths
        This happens after full size images are saved to the filesystem
        '''
        from PIL import Image
        from pyonir import Site
        from pathlib import Path
        raw_img = self.open_image()
        if sizes is None:
            sizes = [Site.THUMBNAIL_DEFAULT]
        try:
            for dimensions in sizes:
                width, height = dimensions
                # self._sizes.append(dimensions)
                img = raw_img.resize((width, height), Image.Resampling.BICUBIC)
                file_name = f'{self.file_name}--{width}x{height}'
                img_dirpath = os.path.dirname(self.file_path)
                Path(img_dirpath).mkdir(parents=True, exist_ok=True)
                filepath = os.path.join(img_dirpath, Site.UPLOADS_THUMBNAIL_DIRNAME, file_name + '.' + self.file_ext)
                if not os.path.exists(filepath): img.save(filepath)
        except Exception as e:
            raise

    @staticmethod
    def decode_filename(encoded_filename: str) -> Optional[dict]:
        """ Reverse of encode_filename. """
        from urllib.parse import parse_qs
        import base64

        try:
            # restore padding
            padding = "=" * (-len(encoded_filename) % 4)
            encoded_filename = encoded_filename.replace("_", ".") + padding
            raw = base64.urlsafe_b64decode(encoded_filename.encode()).decode()
            parsed = parse_qs(raw)
            return parsed
        except Exception as e:
            return None

    @staticmethod
    def encode_filename(file_path: str, meta_data: dict = None) -> str:
        """
        Build filename as url encoded string, then Base64 encode (URL-safe, no '.' in output).
        """
        from urllib.parse import urlencode
        from datetime import datetime
        import base64

        file_name, file_ext = os.path.splitext(os.path.basename(file_path))
        created_date = int(datetime.now().timestamp())
        raw = urlencode(meta_data) if meta_data else f'name={file_name}&ext={file_ext}&created_date={created_date}'
        # URL-safe base64 (no + or /), strip padding '='
        b64 = base64.urlsafe_b64encode(raw.encode()).decode().rstrip("=")
        return b64+file_ext

    @staticmethod
    async def save_upload(file, img_folder_abspath) -> str:
        """Saves base64 file contents into file system"""
        from pathlib import Path
        file_name, file_ext = os.path.splitext(file.filename)
        new_dir_path = Path(img_folder_abspath)
        new_dir_path.mkdir(parents=True, exist_ok=True)
        new_file_path = os.path.join(img_folder_abspath, file_name + file_ext)
        file_contents = await file.read()
        if not file_contents: return ''
        with open(str(new_file_path), 'wb') as f:
            f.write(file_contents)
        return new_file_path

    @staticmethod
    def get_media_data(media_file_path: str):
        from pymediainfo import MediaInfo
        media_info = MediaInfo.parse(media_file_path)
        media_track_file = media_info.tracks.pop(0)
        created_on = media_track_file.file_creation_date
        for track in media_info.tracks:
            if track.track_type == "Image":
                return {
                    "name": media_track_file.file_name,
                    "created_on": created_on,
                    "width": track.width,
                    "height": track.height,
                }
            if track.track_type == "Audio":
                dur = track.duration / 1000 if track.duration else None # ms → seconds
                return {
                    "codec": track.codec,
                    "duration": dur,
                    "bit_rate": track.bit_rate,
                    "channels": track.channel_s,
                    "sampling_rate": track.sampling_rate,
                }
            if track.track_type == "Video":
                return {
                    "codec": track.codec,
                    "duration": track.duration / 1000 if track.duration else None,  # ms → seconds
                    "width": track.width,
                    "height": track.height,
                    "frame_rate": track.frame_rate,
                    "bit_rate": track.bit_rate,
                }


class MediaManager:
    """Manage audio, video, and image documents."""
    default_media_dirname = 'media'

    def __init__(self, app: 'BaseApp'):
        self.app = app
        self.supported_formats = {ImageFormat.JPG, ImageFormat.PNG, VideoFormat.MP4, AudioFormat.MP3}
        self._storage_dirpath: str = os.path.join(app.contents_dirpath, self.default_media_dirname)
        """Location on fs to save file uploads"""

    @property
    def storage_dirpath(self) -> str: return self._storage_dirpath

    def is_supported(self, ext: str) -> bool:
        """Check if the media file has a supported format."""
        ext = ext.lstrip('.').lower()
        return ext in {fmt.value for fmt in self.supported_formats}

    def add_supported_format(self, fmt: Union[ImageFormat, AudioFormat, VideoFormat, None]):
        """Add a supported media format."""
        self.supported_formats.add(fmt)

    @staticmethod
    def media_type(ext: str) -> Optional[str]:
        """Return the media type based on file extension."""
        ext = ext.lstrip('.').lower()

        if ext in (f.value for f in AudioFormat):
            return "audio"
        elif ext in (f.value for f in VideoFormat):
            return "video"
        elif ext in (f.value for f in ImageFormat):
            return "image"
        return None

    def set_storage_dirpath(self, storage_dirpath):
        self._storage_dirpath = storage_dirpath
        return self

    def close(self):
        """Closes any open connections by resetting storage path"""
        self._storage_dirpath = os.path.join(self.app.contents_dirpath, self.default_media_dirname)

    def get_media(self, file_id: str) -> BaseMedia:
        """Retrieves user paginated media files"""
        mpath = os.path.join(self.storage_dirpath, file_id)
        mfile = BaseMedia(mpath)
        return mfile

    def get_medias(self, file_type: str) -> list[BaseMedia]:
        """Retrieves user paginated media files"""
        files = BaseFSQuery(self.storage_dirpath, model=BaseMedia, force_all=True)
        return list(files)

    def delete_media(self, media_id: str) -> bool:
        """Delete file by ID. Returns True if deleted."""
        from pathlib import Path
        path = os.path.join(self.storage_dirpath, media_id)
        if not Path(path).exists(): return False
        Path(path).unlink()
        return True

    # --- General Uploading ---
    async def upload(self, request: PyonirRequest, directory_name: str = None, file_name: str = None, limit: int = None) -> \
            list[BaseMedia]:
        """Uploads a resource into specified directory
        :param request: PyonirRequest instance
        :param directory_name: directory name
        :param file_name: strict file name for resource
        :param limit: maximum number of files to upload
        """
        resource_files = []
        for file in request.files:
            if limit and len(resource_files) == limit: break
            if file_name:
                file._filename = file_name
            directory_name = directory_name or self.media_type(file.ext)
            media_file: BaseMedia = await self._upload_bytes(file, directory_name=directory_name or self.default_media_dirname)
            if media_file:
                resource_files.append(media_file)
        return resource_files

    async def _upload_bytes(self, file: UploadFile, directory_name: str = None, meta_data: dict = None) -> Optional[BaseMedia]:
        """
        Save an uploaded video file to disk and return its filename.
        or upload a video to Cloudflare R2 and return the object key.
        """
        from pathlib import Path
        filename = file.filename
        _strict_name = getattr(file, '_filename', None)
        if not filename: return None
        resource_id = [f"{directory_name.strip()}", f"{_strict_name.strip() if _strict_name else filename}"]
        path = os.path.join(self.storage_dirpath, *resource_id)
        Path(os.path.dirname(path)).mkdir(parents=True, exist_ok=True)
        with open(path, "wb") as buffer:
            while chunk := await file.read(1024 * 1024):  # 1MB chunks
                buffer.write(chunk)
        file_media = BaseMedia(path=path)
        return file_media






# import asyncio
# import os, json, inspect
# from typing import Optional, Callable, Tuple, AsyncGenerator, List, get_type_hints
# from starlette.applications import Starlette
# from starlette.websockets import WebSocket, WebSocketState, WebSocketDisconnect
# from pyonir.core import PyonirApp, PyonirRequest
# from pyonir.pyonir_types import PyonirServer, PyonirHooks, AppEndpoint, PyonirRouters, \
#     PyonirRestResponse
# from pyonir.utilities import create_file, get_attr, cls_mapper
#
# TEXT_RES = 'text/html'
# JSON_RES = 'application/json'
# EVENT_RES = 'text/event-stream'
#
# ConnClients = {}
#
# class PyonirServer(Starlette):
#
#     def __init__(self, app: PyonirApp):
#         # super().__init__()
#         self.is_active = False
#         self.app = app
#         self.ws_routes = []
#         self.sse_routes = []
#         self.auth_routes = []
#         self.endpoints = []
#         self.url_map = {}
#         self.resolvers = {}
#         self.services = {}
#         self.request: Optional[PyonirRequest] = None
#
#     @property
#     def allowed_hosts(self):
#         """Returns a list of domains allowed to access the web application"""
#         return ['localhost', '*.localhost']
#
#     def run_uvicorn_server(self, endpoints: PyonirRouters = None):
#         """Starts the webserver"""
#         import uvicorn, sys
#
#         # """Uvicorn web server configurations"""
#         uvicorn_options = {
#             "port": self.app.port,
#             "host": self.app.host
#         }
#         if self.app.is_secure:
#             uvicorn_options["ssl_keyfile"] = self.app.ssl_key_file
#             uvicorn_options["ssl_certfile"] = self.app.ssl_cert_file
#         if not self.app.is_dev:
#             uvicorn_options['uds'] = self.app.unix_socket_filepath
#
#         # Initialize routers
#         if endpoints: init_app_endpoints(endpoints)
#         init_pyonir_endpoints(self.app)
#         print(f"/************** ASGI APP SERVER RUNNING on {'http' if self.app.is_dev else 'sock'} ****************/")
#         print(f"\
#         \n\t- App env: {'DEV' if self.app.is_dev else 'PROD'}\
#         \n\t- App name: {self.app.name}\
#         \n\t- App domain: {self.app.domain}\
#         \n\t- App host: {self.app.host}\
#         \n\t- App port: {self.app.port}\
#         \n\t- App sock: {self.app.unix_socket_filepath}\
#         \n\t- App ssl_key: {self.app.ssl_key_file}\
#         \n\t- App ssl_cert: {self.app.ssl_cert_file}\
#         \n\t- App Server: Uvicorn \
#         \n\t- NGINX config: {self.app.nginx_config_filepath} \
#         \n\t- System Version: {sys.version_info}")
#         self.app.run_plugins(PyonirHooks.AFTER_INIT)
#         # self.app.run_hooks(PyonirHooks.AFTER_INIT)
#         uvicorn.run(self.app.server, **uvicorn_options)
#
#     def create_router(self, endpoints: 'Endpoints'):
#         """A collection (or group) of routes, usually organized by feature or resource, and often mounted under"""
#         for endpoint, routes in endpoints:
#             for path, func, methods, *opts in routes:
#                 self.add_route(func, path=f'{endpoint}{path}', methods=methods, *opts)
#                 pass
#
#     def create_route(self, dec_func: Callable,
#                        path: str = '',
#                        methods=None,
#                        auth: bool = None,
#                        ws: bool = None,
#                        sse: bool = None,
#                        static_path: str = None) -> Optional[Callable]:
#         """A mapping of an HTTP method and an endpoint to a handler function."""
#         from pyonir import Site
#
#         is_async = inspect.iscoroutinefunction(dec_func) if dec_func else False
#         is_asyncgen = inspect.isasyncgenfunction(dec_func) if dec_func else False
#         if methods == '*':
#             methods = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE']
#         if methods is None:
#             methods = ['GET']
#
#         if static_path:
#             static_route = self.serve_static(static_path)
#             self.mount(path, static_route)
#             self.url_map[path.lstrip('/')
#             .replace('/', '_')] = {'path': path, 'dir': static_path, 'exists': os.path.exists(static_path)}
#             return None
#
#         route_name = dec_func.__name__
#         docs = dec_func.__doc__
#         route_path = path.split('/{')[0]
#         name = route_name
#         endpoint_route = path.split('/', 1)[0]
#         # req_models = Site.server.url_map.get(route_name, {}).get('models') or {}
#
#         new_route = {
#             "doc": docs,
#             "endpoint": endpoint_route,
#             "params": {k: v.__name__ for k, v in dec_func.__annotations__.items()},
#             "route": path,  # has regex pattern
#             "path": route_path,
#             "methods": methods,
#             # "models": req_models,
#             "name": name,
#             "auth": auth,
#             "sse": sse,
#             "ws": ws,
#             "async": is_async,
#             "async_gen": is_asyncgen,
#         }
#         # Add route path into categories
#         self.endpoints.append(f"{endpoint_route}{route_path}")
#         self.url_map[name] = new_route
#         if sse:
#             self.sse_routes.append(f"{endpoint_route}{route_path}")
#         if auth:
#             self.auth_routes.append(f"{endpoint_route}{route_path}")
#         if ws:
#             self.ws_routes.append(f"{endpoint_route}{route_path}")
#             return self.add_websocket_route(path, dec_func, dec_func.__name__)
#
#         async def dec_wrapper(star_req):
#             from pyonir.core import PyonirRequest
#             pyonir_request = PyonirRequest(star_req, Site)
#             return await self.build_response(pyonir_request, dec_func)
#
#         dec_wrapper.__name__ = dec_func.__name__
#         dec_wrapper.__doc__ = dec_func.__doc__
#         self.add_route(path, dec_wrapper, methods=methods)
#
#     def initialize_starlette(self) -> PyonirServer:
#         """Setup Starlette web server"""
#         super().__init__()
#         from starlette_wtf import CSRFProtectMiddleware
#         from starlette.middleware.sessions import SessionMiddleware
#         from starlette.middleware.trustedhost import TrustedHostMiddleware
#         # from starlette.middleware.gzip import GZipMiddleware
#
#         # allowed_hosts = ['localhost', '*.localhost']
#         self.add_middleware(TrustedHostMiddleware)
#         # star_app.add_middleware(GZipMiddleware, minimum_size=500)
#         self.add_middleware(SessionMiddleware,
#                                 https_only=False,
#                                 secret_key=self.app.SECRET_SAUCE,
#                                 session_cookie=self.app.SESSION_KEY,
#                                 same_site='lax'
#                                 )
#         self.add_middleware(CSRFProtectMiddleware, csrf_secret=self.app.SECRET_SAUCE)
#
#     @staticmethod
#     async def build_response(pyonir_request: PyonirRequest, dec_func: callable):
#         from pyonir import Site
#         from pyonir.models.auth import Auth
#
#         star_req = pyonir_request.server_request
#         Site.server.request = pyonir_request
#         default_system_router = dec_func.__name__ == 'pyonir_index'
#         # Serve static urls
#         if pyonir_request.is_static:
#             return serve_static(Site, pyonir_request)
#
#         # Update template globals
#         Site.TemplateEnvironment.globals['request'] = pyonir_request
#
#         # Preprocess request form data
#         await pyonir_request.process_request_data()
#         # Init Auth handler
#         pyonir_request.auth = Auth(pyonir_request, Site)
#
#
#         # Process request context i.e plugin vs app
#         pyonir_request.get_context()
#         # Process file associated with request
#         app_ctx, req_file = pyonir_request.resolve_request_to_file(Site)
#         pyonir_request.file = req_file
#         pyonir_request.app_ctx_name = app_ctx.name
#
#         # Preprocess routes or resolver endpoint from file
#         req_file = await app_ctx.virtual_router(pyonir_request)
#         await req_file.process_resolver(pyonir_request)
#
#         route_func = dec_func if not callable(req_file.resolver) else req_file.resolver
#
#         # Check resolver route security
#         security = pyonir_request.auth.security.check(pyonir_request.auth)
#         if security and not security.accepted:
#             return Site.server.serve_redirect(security.redirect_to or '/')
#
#         # Get router endpoint from map
#         if callable(route_func):
#             is_async = inspect.iscoroutinefunction(route_func)
#
#             default_args = dict(get_type_hints(route_func))
#             default_args.update(**pyonir_request.path_params)
#             default_args.update(**pyonir_request.query_params.__dict__)
#             default_args.update(**pyonir_request.form)
#             args = cls_mapper(default_args, route_func, from_request=pyonir_request)
#             pyonir_request.router_args = args
#
#             # Resolve route decorator methods
#             pyonir_request.server_response = await route_func(**args) if is_async else route_func(**args)
#
#         else:
#             pyonir_request.server_response = req_file.resolver or req_file.data
#
#         # Perform redirects
#         if pyonir_request.redirect_to:
#             return Site.server.serve_redirect(pyonir_request.redirect_to)
#
#         # Derive status code
#         pyonir_request.derive_status_code(Site.server.url_map.get(route_func.__name__ if route_func else '') and not default_system_router)
#
#         # Execute plugins hooks initial request
#         await Site.run_async_plugins(PyonirHooks.ON_REQUEST, pyonir_request)
#         # await Site.run_async_hooks(PyonirHooks.ON_REQUEST, pyonir_request)
#
#         # Finalize response output
#         is_auth_res = isinstance(pyonir_request.server_response, PyonirRestResponse)
#         pyonir_response = pyonir_request.server_response if is_auth_res else PyonirRestResponse(status_code=pyonir_request.status_code)
#         if not is_auth_res:
#             if pyonir_request.type == JSON_RES:
#                 pyonir_response.set_json(pyonir_request.server_response or pyonir_request.file.data)
#             else:
#                 pyonir_response.set_html(req_file.output_html(pyonir_request))
#
#         # Set headers
#         pyonir_response.set_media(pyonir_request.type)
#         pyonir_response.set_header('Server', 'Pyonir Web Framework')
#         pyonir_response.set_header("Cache-Control", "no-store, no-cache, must-revalidate, max-age=0")
#         pyonir_response.set_header("Pragma", "no-cache")
#         pyonir_response.set_header("Expires", "0")
#
#         # Generate response
#         pyonir_response.set_server_response()
#         return pyonir_response.render()
#
#
#     @staticmethod
#     def response_renderer(value, media_type):
#         from starlette.responses import Response, StreamingResponse
#         if media_type == EVENT_RES:
#             return StreamingResponse(content=value, media_type=media_type)
#         return Response(content=value, media_type=media_type)
#
#     @staticmethod
#     def serve_redirect(url: str, code=302):
#         from starlette.responses import RedirectResponse
#         res = RedirectResponse(url=url.strip(), status_code=code)
#         res.headers["Cache-Control"] = "no-store, no-cache, must-revalidate, max-age=0"
#         res.headers["Pragma"] = "no-cache"
#         res.headers["Expires"] = "0"
#         return res
#
#     @staticmethod
#     def serve_static(assets_dir: str):
#         """Creates a route for serving static files"""
#         from starlette.staticfiles import StaticFiles
#         return StaticFiles(directory=assets_dir)
#
#
# async def pyonir_ws_handler(websocket: WebSocket):
#     """ws connection handler"""
#     from pyonir.utilities import generate_id
#
#     async def get_data(ws: WebSocket):
#         assert ws.application_state == WebSocketState.CONNECTED and ws.client_state == WebSocketState.CONNECTED
#         wsdata = await ws.receive()
#
#         if wsdata.get('text'):
#             wsdata = wsdata['text']
#             swsdata = json.loads(wsdata)
#             swsdata['value'] = swsdata.get('value')
#             wsdata = json.dumps(swsdata)
#         elif wsdata.get('bytes'):
#             wsdata = wsdata['bytes'].decode('utf-8')
#
#         return wsdata
#
#     async def broadcast(message: str, ws_id: str = None):
#         for id, ws in ConnClients.items():
#             if active_id == id and hasattr(ws, 'send_text'): continue
#             await ws.send_text(message)
#
#     async def on_disconnect(websocket: WebSocket):
#         del ConnClients[active_id]
#         client_data.update({"action": "ON_DISCONNECTED", "id": active_id})
#         await broadcast(json.dumps(client_data))
#
#     async def on_connect(websocket: WebSocket):
#         client_data.update({"action": "ON_CONNECTED", "id": active_id})
#         await websocket.send_text(json.dumps(client_data))
#
#     active_id = generate_id()
#     client_data = {}
#     await websocket.accept()  # Accept the WebSocket connection
#     print("WebSocket connection established!")
#     ConnClients[active_id] = websocket
#     await on_connect(websocket)
#     try:
#         while websocket.client_state == WebSocketState.CONNECTED:
#             # Wait for a message from the client
#             data = await get_data(websocket)
#             print(f"Received from client: {data}")
#             # Respond to the client
#             await broadcast(data)
#         await on_disconnect(data)
#     except Exception as e:
#         del ConnClients[active_id]
#         print(f"WebSocket connection closed: {active_id}")
#
#
# async def pyonir_sse_handler(request: PyonirRequest) -> AsyncGenerator:
#     """Handles sse web request by pyonir"""
#     from pyonir.utilities import generate_id
#     request.type = EVENT_RES  # assign the appropriate streaming headers
#     # set sse client
#     event = get_attr(request.query_params, 'event')
#     retry = get_attr(request.query_params, 'retry') or 1000
#     close_id = get_attr(request.query_params, 'close')
#     interval = 1  # time between events
#     client_id = get_attr(request.query_params, 'id') or request.headers.get('user-agent')
#     client_id += f"{client_id}{generate_id()}"
#     if close_id and ConnClients.get(close_id):
#         del ConnClients[close_id]
#         return
#     last_client = ConnClients.get(client_id, {
#         "retry": retry,
#         "event": event,
#         "id": client_id,
#         "data": {
#             "time": 0
#         },
#     })
#     # register new client
#     if not ConnClients.get(client_id):
#         ConnClients[client_id] = last_client
#
#     while True:
#         last_client["data"]["time"] = last_client["data"]["time"] + 1
#         is_disconnected = await request.server_request.is_disconnected()
#         if is_disconnected or close_id:
#             del ConnClients[client_id]
#             break
#         await asyncio.sleep(interval)  # Wait for 5 seconds before sending the next message
#         res = process_sse(last_client)
#         yield res
#
#
# async def pyonir_docs_handler(request: PyonirRequest):
#     """Documentation for every endpoint by pyonir"""
#     return {"routes": request.server_request.app.url_map, "configs": request.auth.app._env}
#
#
# def pyonir_index(request: PyonirRequest):
#     """Catch all routes for all web request"""
#     pass
#
#
# def setup_starlette_server(iapp: PyonirApp) -> PyonirServer:
#     """Setup Starlette web server"""
#
#     from starlette_wtf import CSRFProtectMiddleware
#     from starlette.middleware.sessions import SessionMiddleware
#
#     from starlette.middleware.trustedhost import TrustedHostMiddleware
#     from starlette.middleware.gzip import GZipMiddleware
#     from starlette.responses import Response, StreamingResponse
#     from starlette.routing import Router, Route
#
#     def render_response(value, media_type):
#         if media_type == EVENT_RES:
#             return StreamingResponse(content=value, media_type=media_type)
#         return Response(content=value, media_type=media_type)
#
#     def redirect(url, code=302):
#         from starlette.responses import RedirectResponse
#         res = RedirectResponse(url=url.strip(), status_code=code)
#         res.headers["Cache-Control"] = "no-store, no-cache, must-revalidate, max-age=0"
#         res.headers["Pragma"] = "no-cache"
#         res.headers["Expires"] = "0"
#         return res
#
#     def get_staticroute(assets_dir):
#         from starlette.staticfiles import StaticFiles
#         return StaticFiles(directory=assets_dir)
#
#     allowed_hosts = ['localhost', '*.localhost']
#     star_app = PyonirServer()  # inherits from Starlette
#     star_app.add_middleware(TrustedHostMiddleware)
#     # star_app.add_middleware(GZipMiddleware, minimum_size=500)
#     star_app.add_middleware(SessionMiddleware,
#                             https_only=False,
#                             secret_key=iapp.SECRET_SAUCE,
#                             session_cookie=iapp.SESSION_KEY,
#                             same_site='lax'
#                             )
#     star_app.add_middleware(CSRFProtectMiddleware, csrf_secret=iapp.SECRET_SAUCE)
#
#     # Interface properties required for pyonir to process web request
#     setattr(star_app, 'response_renderer', render_response)
#     setattr(star_app, 'serve_redirect', redirect)
#     setattr(star_app, 'create_endpoint', Router)
#     setattr(star_app, 'create_route', Route)
#     setattr(star_app, 'serve_static', get_staticroute)
#
#     return star_app
#
#
# def init_app_endpoints(endpoints: 'Endpoints'):
#     for endpoint, routes in endpoints:
#         for path, func, methods, *opts in routes:
#             _add_route(func, f'{endpoint}{path}', methods, *opts)
#             pass
#
#
# def init_pyonir_endpoints(app: PyonirApp):
#     app.server.create_route(pyonir_ws_handler, "/sysws", ws=True)
#     app.server.create_route(pyonir_index, "/", methods='*')
#     app.server.create_route(pyonir_index, "/{path:path}", methods='*')
#
# def process_sse(data: dict) -> str:
#     """Formats a string and an event name in order to follow the event stream convention.
#     'event: Jackson 5\\ndata: {"abc": 123}\\n\\n'
#     """
#     sse_payload = ""
#     for key, val in data.items():
#         val = json.dumps(val) if key == 'data' else val
#         sse_payload += f"{key}: {val}\n"
#     return sse_payload + "\n"
#
# def serve_static(app: PyonirApp, request: PyonirRequest):
#     from starlette.responses import FileResponse, PlainTextResponse
#     from_frontend = request.path.startswith(app.frontend_assets_route)
#     from_public = request.path.startswith(app.public_assets_route)
#     base_path = app.frontend_assets_dirpath if from_frontend else app.public_assets_dirpath if from_public else ''
#     req_path = request.parts[1:] if len(request.parts) > 1 else request.parts
#     path = os.path.join(base_path, *req_path)
#     print('Pyonir: serving static: '+ path)
#     return FileResponse(path, 200) if os.path.exists(path) else PlainTextResponse(f"{request.path} not found",
#                                                                                   status_code=404)
#
# def route(dec_func: Callable,
#           url: str = '',
#           methods=None,
#           auth: bool = None,
#           ws: bool = None,
#           sse: bool = None,
#           static_path: str = None) -> Callable:
#     if dec_func is None and static_path is None:
#         dec_func = pyonir_index
#     return _add_route(dec_func, path=url, methods=methods, auth=auth, ws=ws, sse=sse, static_path=static_path)
#
#
# def _add_route(dec_func: Callable,
#                path: str = '',
#                methods=None,
#                auth: bool = None,
#                ws: bool = None,
#                sse: bool = None,
#                static_path: str = None) -> Callable:
#     """Route decorator"""
#     from pyonir import Site
#
#     is_async = inspect.iscoroutinefunction(dec_func) if dec_func else False
#     is_asyncgen = inspect.isasyncgenfunction(dec_func) if dec_func else False
#     # list_of_args = list(inspect.signature(dec_func).parameters.keys()) if dec_func else None
#     if methods == '*':
#         methods = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE']
#     if methods is None:
#         methods = ['GET']
#
#     if static_path:
#         from starlette.staticfiles import StaticFiles
#         static_route = Site.server.serve_static(static_path)
#         Site.server.mount(path, static_route)
#         Site.server.url_map[path.lstrip('/')
#         .replace('/', '_')] = {'path': path, 'dir': static_path, 'exists': os.path.exists(static_path)}
#         return
#
#     route_name = dec_func.__name__
#     docs = dec_func.__doc__
#     route_path = path.split('/{')[0]
#     name = route_name
#     endpoint_route = path.split('/', 1)[0]
#     # req_models = Site.server.url_map.get(route_name, {}).get('models') or {}
#
#     new_route = {
#         "doc": docs,
#         "endpoint": endpoint_route,
#         "params": {k: v.__name__ for k, v in dec_func.__annotations__.items()},
#         "route": path,  # has regex pattern
#         "path": route_path,
#         "methods": methods,
#         # "models": req_models,
#         "name": name,
#         "auth": auth,
#         "sse": sse,
#         "ws": ws,
#         "async": is_async,
#         "async_gen": is_asyncgen,
#     }
#     # Add route path into categories
#     Site.server.endpoints.append(f"{endpoint_route}{route_path}")
#     Site.server.url_map[name] = new_route
#     if sse:
#         Site.server.sse_routes.append(f"{endpoint_route}{route_path}")
#     if auth:
#         Site.server.auth_routes.append(f"{endpoint_route}{route_path}")
#     if ws:
#         Site.server.ws_routes.append(f"{endpoint_route}{route_path}")
#         return Site.server.add_websocket_route(path, dec_func, dec_func.__name__)
#
#     async def dec_wrapper(star_req):
#         from pyonir.core import PyonirRequest
#         pyonir_request = PyonirRequest(star_req, Site)
#         return await build_response(pyonir_request, dec_func)
#
#     dec_wrapper.__name__ = dec_func.__name__
#     dec_wrapper.__doc__ = dec_func.__doc__
#     Site.server.add_route(path, dec_wrapper, methods=methods)
#
#
# async def __build_response(pyonir_request: PyonirRequest, dec_func: callable):
#     from pyonir import Site
#     from pyonir.models.auth import Auth
#
#     star_req = pyonir_request.server_request
#     Site.server.request = pyonir_request
#     default_system_router = dec_func.__name__ == 'pyonir_index'
#     # Serve static urls
#     if pyonir_request.is_static:
#         return serve_static(Site, pyonir_request)
#
#     # Update template globals
#     Site.TemplateEnvironment.globals['request'] = pyonir_request
#
#     # Preprocess request form data
#     await pyonir_request.process_request_data()
#     # Init Auth handler
#     pyonir_request.auth = Auth(pyonir_request, Site)
#
#
#     # Process file associated with request
#     app_ctx, req_file = pyonir_request.resolve_request_to_file(star_req.url.path, Site)
#     pyonir_request.file = req_file
#     pyonir_request.app_ctx_name = app_ctx.name
#
#     # Preprocess routes or resolver endpoint from file
#     req_file = await Site.virtual_router(pyonir_request)
#     await req_file.process_resolver(pyonir_request)
#
#     route_func = dec_func if not callable(req_file.resolver) else req_file.resolver
#
#     # Check resolver route security
#     security = pyonir_request.auth.security.check(pyonir_request.auth)
#     if security and not security.accepted:
#         return Site.server.serve_redirect(security.redirect_to or '/')
#
#     # Get router endpoint from map
#     if callable(route_func):
#         is_async = inspect.iscoroutinefunction(route_func)
#
#         default_args = dict(get_type_hints(route_func))
#         default_args.update(**pyonir_request.path_params)
#         default_args.update(**pyonir_request.query_params.__dict__)
#         default_args.update(**pyonir_request.form)
#         args = cls_mapper(default_args, route_func, from_request=pyonir_request)
#         pyonir_request.router_args = args
#
#         # Resolve route decorator methods
#         pyonir_request.server_response = await route_func(**args) if is_async else route_func(**args)
#
#     else:
#         pyonir_request.server_response = req_file.resolver or req_file.data
#
#     # Perform redirects
#     if pyonir_request.redirect_to:
#         return Site.server.serve_redirect(pyonir_request.redirect_to)
#
#     # Derive status code
#     pyonir_request.derive_status_code(Site.server.url_map.get(route_func.__name__ if route_func else '') and not default_system_router)
#
#     # Execute plugins hooks initial request
#     await Site.run_async_plugins(PyonirHooks.ON_REQUEST, pyonir_request)
#     # await Site.run_async_hooks(PyonirHooks.ON_REQUEST, pyonir_request)
#
#     # Finalize response output
#     is_auth_res = isinstance(pyonir_request.server_response, PyonirRestResponse)
#     pyonir_response = pyonir_request.server_response if is_auth_res else PyonirRestResponse(status_code=pyonir_request.status_code)
#     if not is_auth_res:
#         if pyonir_request.type == JSON_RES:
#             pyonir_response.set_json(pyonir_request.server_response or pyonir_request.file.data)
#         else:
#             pyonir_response.set_html(req_file.output_html(pyonir_request))
#
#     # Set headers
#     pyonir_response.set_media(pyonir_request.type)
#     pyonir_response.set_header('Server', 'Pyonir Web Framework')
#     pyonir_response.set_header("Cache-Control", "no-store, no-cache, must-revalidate, max-age=0")
#     pyonir_response.set_header("Pragma", "no-cache")
#     pyonir_response.set_header("Expires", "0")
#
#     # Generate response
#     pyonir_response.set_server_response()
#     return pyonir_response.render()
#
# def generate_nginx_conf(app: PyonirApp) -> bool:
#     """Generates a NGINX conf file based on App configurations"""
#
#     nginx_conf = app.TemplateEnvironment.get_template("nginx.jinja.conf") \
#         .render(
#         app_name=app.name,
#         app_name_id=app.name.replace(' ', '_').lower(),
#         domain=app.domain,
#         is_secure=app.is_secure,
#         ssl_cert_file=app.ssl_cert_file,
#         ssl_key_file=app.ssl_key_file,
#         site_dirpath=app.app_dirpath,
#         site_logs_dirpath=app.logs_dirpath,
#         app_socket_filepath=app.unix_socket_filepath,
#         app_ignore_logs=f"{app.PUBLIC_ASSETS_DIRNAME}|{app.UPLOADS_DIRNAME}|{app.FRONTEND_ASSETS_DIRNAME}",
#         frontend_assets_route=app.frontend_assets_route,
#         frontend_assets_dirpath=app.frontend_assets_dirpath,
#         public_assets_route=app.public_assets_route,
#         public_assets_dirpath=app.public_assets_dirpath,
#         site_uploads_route=app.uploads_route,
#         site_uploads_dirpath=app.uploads_dirpath,
#         site_ssg_dirpath=app.ssg_dirpath,
#         custom_nginx_locations=get_attr(app._env, 'nginx_locations')
#     )
#
#     return create_file(app.nginx_config_filepath, nginx_conf, False)
#



import pytest, os
from typing import Optional, Union, List, Dict

from pyonir.models.mapper import cls_mapper
from pyonir.models.parser import DeserializeFile
from pyonir.utilities import parse_query_model_to_object


# ==== Sample classes to map into ====

class Address:
    street: str
    zip_code: Optional[int]

    def __init__(self, street: str, zip_code: Optional[int] = None):
        self.street = street
        self.zip_code = zip_code

class User:
    id: int
    name: str
    email: Optional[str]
    address: Optional[Address]
    tags: List[str]
    meta: Dict[str, Union[str, int]]

    def __init__(self, id: int, name: str, email: Optional[str],
                 address: Optional[Address], tags: List[str], meta: Dict[str, Union[str, int]]):
        self.id = id
        self.name = name
        self.email = email
        self.address = address
        self.tags = tags
        self.meta = meta

class Article:
    """a media post."""

    _orm_options = {'mapper': {'id': 'file_name', 'caption': 'content'},'frozen': True}
    """dict: Internal mapping of model fields to source attributes."""

    def __init__(self, caption: str = None, title: str = None, alt: str = None):
        from datetime import datetime
        import uuid
        self.title: str = title
        self.caption: str = caption
        self.alt: str = alt
        self.id: str = uuid.uuid4().hex
        self.created_on: datetime = datetime.now()
        self.last_updated: datetime = datetime.now()

generic_model = parse_query_model_to_object('title,url,author,date:file_created_on')
article_filepath = os.path.join(os.path.dirname(__file__), 'contents', 'article.md')
# ==== Tests ====

def test_parsely_to_custom_mapping():
    obj = DeserializeFile(article_filepath)
    article = cls_mapper(obj, Article)
    assert isinstance(article, Article)
    assert article.id is not None
    assert article.caption == obj.data['content']


def test_no_hint_mapping():
    obj = {"title": "hunter", "author": "Alice", "url": "/foo", "date": None}
    genmodel = cls_mapper(obj, generic_model)
    assert genmodel.author == "Alice"
    assert genmodel.url == '/foo'

def test_scalar_mapping():
    obj = {"id": "123", "name": "Alice", "email": None, "address": None, "tags": [], "meta": {}}
    user = cls_mapper(obj, User)
    assert isinstance(user.id, int)
    assert user.id == 123
    assert user.name == "Alice"
    assert user.email is None

def test_optional_mapping():
    obj = {"id": 1, "name": "Bob", "email": "bob@test.com", "address": None, "tags": [], "meta": {}}
    user = cls_mapper(obj, User)
    assert user.email == "bob@test.com"
    obj2 = {"id": 2, "name": "Charlie", "email": None, "address": None, "tags": [], "meta": {}}
    user2 = cls_mapper(obj2, User)
    assert user2.email is None

def test_nested_object():
    obj = {
        "id": 10, "name": "Diana", "email": "diana@test.com",
        "address": {"street": "Main St", "zip_code": "90210"},
        "tags": ["admin", "staff"],
        "meta": {"age": "30", "score": 95}
    }
    user = cls_mapper(obj, User)
    assert isinstance(user.address, Address)
    assert user.address.street == "Main St"
    assert isinstance(user.address.zip_code, int)
    assert user.address.zip_code == 90210
    assert user.meta["score"] == 95  # int conversion
    assert user.meta["age"] == "30"  # str preserved

def test_list_mapping():
    obj = {
        "id": 20, "name": "Eva", "email": None,
        "address": None,
        "tags": ["one", "two"],
        "meta": {}
    }
    user = cls_mapper(obj, User)
    assert isinstance(user.tags, list)
    assert user.tags == ["one", "two"]

def test_dict_mapping_with_union():
    obj = {
        "id": 30, "name": "Frank", "email": None,
        "address": None,
        "tags": [],
        "meta": {"age": 42, "nickname": "franky"}
    }
    user = cls_mapper(obj, User)
    assert isinstance(user.meta["age"], int)
    assert isinstance(user.meta["nickname"], str)



import os, json

def readJSON(abs_path_to_json_file):
	jsondata = {}
	if not abs_path_to_json_file.endswith(".json"): return f"{abs_path_to_json_file} is not a valid json file"
	try:
		with open(abs_path_to_json_file) as json_file:
			jsondata = json.load(json_file)
	except Exception as e:
		jsondata = f"{type(e).__name__} error occurred when attempting to convert {abs_path_to_json_file}: {e}"
	return jsondata



from __future__ import annotations
import os
from dataclasses import dataclass, field
from typing import Optional, Union, Tuple, Callable, get_type_hints, OrderedDict, TYPE_CHECKING, Any

from starlette.applications import Starlette
from starlette.requests import Request as StarletteRequest

from pyonir.models.app import BaseApp
from pyonir.models.parser import DeserializeFile
from pyonir.pyonir_types import PyonirRouters, PyonirHooks
from pyonir.utilities import dict_to_class

TEXT_RES: str = 'text/html'
JSON_RES: str = 'application/json'
EVENT_RES: str = 'text/event-stream'

# Environments
LOCAL_ENV:str = 'LOCAL'
DEV_ENV:str = 'DEV'
PROD_ENV:str = 'PROD'




class BaseRequest:
    PAGINATE_LIMIT: int = 6

    def __init__(self, server_request: Optional[StarletteRequest], app: BaseApp):
        from pyonir.utilities import get_attr
        from pyonir.models.auth import Auth

        self.server_response = None
        self.file: Optional[DeserializeFile] = None
        self.server_request: StarletteRequest = server_request
        self.raw_path = "/".join(str(self.server_request.url).split(str(self.server_request.base_url))) if server_request else ''
        self.method = self.server_request.method if server_request else 'GET'
        self.path = self.server_request.url.path if server_request else '/'
        self.path_params = dict_to_class(self.server_request.path_params,'path_params') if server_request else {}
        self.url = f"{self.path}" if server_request else {}
        self.slug = self.path.lstrip('/').rstrip('/')
        self.query_params = self.get_params(self.server_request.url.query) if server_request else {}
        self.parts = self.slug.split('/') if self.slug else []
        self.limit = get_attr(self.query_params, 'limit', self.PAGINATE_LIMIT)
        self.model = get_attr(self.query_params, 'model')
        self.is_home = (self.slug == '')
        self.is_api = self.parts and self.parts[0] == app.API_DIRNAME
        self.is_static = bool(list(os.path.splitext(self.path)).pop()) if server_request else False
        self.form = {}
        self.files = []
        # self.ip = self.server_request.client.host if server_request else ''
        self.host = str(self.server_request.base_url).rstrip('/') if server_request else app.host
        self.protocol = self.server_request.scope.get('type') + "://" if server_request else app.protocol
        self.headers = self.process_header(self.server_request.headers) if server_request else {}
        self.browser = self.headers.get('user-agent', '').split('/').pop(0) if self.headers else "UnknownAgent"
        if self.slug.startswith('api'): self.headers['accept'] = JSON_RES
        self.type: Union[TEXT_RES, JSON_RES, EVENT_RES] = self.headers.get('accept')
        self.status_code: int = 200
        self._app_ctx_ref = None
        self.app = app
        self.auth: Optional[Auth] = None
        self.flashes: dict = self.get_flash_messages() if server_request else {}
        if server_request:
            self.server_request.session['previous_url'] = self.headers.get('referer', '')

    @property
    def app_ctx_ref(self):
        return self._app_ctx_ref or self.app

    @property
    def session_token(self):
        """Returns active csrf token for user session"""
        if self.server_request and self.server_request.session:
            return self.server_request.session.get('csrf_token')

    @property
    def previous_url(self) -> str:
        return self.server_request.session.pop('previous_url')

    @property
    def redirect_to(self):
        """Returns the redirect URL from the request form data"""
        file_redirect = self.file.data.get('redirect_to', self.file.data.get('redirect'))
        return self.form.get('redirect_to', self.form.get('redirect', file_redirect))

    def process_resolver(self):
        from pyonir import Site
        from pyonir.models.parser import FileStatuses
        req_file = self.file
        resolver_obj = req_file.data.get(req_file.RESOLVER_KEY, {})
        resolver_action = resolver_obj.get(self.method)
        # if resolver_obj and not resolver_action:
        #     req_file.status = FileStatuses.FORBIDDEN
        #     self.type = JSON_RES
        if not resolver_action or req_file.resolver is not None: return
        resolver_path = resolver_action.get('call')
        resolver_args = resolver_action.get('args')
        resolver_redirect = resolver_action.get('redirect')

        self.type = resolver_action.get('headers', {}).get('accept', self.type)
        app_plugin = list(filter(lambda p: p.name == resolver_path.split('.')[0], Site.activated_plugins))
        app_plugin = app_plugin[0] if len(app_plugin) else Site
        resolver = app_plugin.reload_resolver(resolver_path)
        if resolver and resolver_args:
            self.form.update(resolver_args)
        if resolver and resolver_redirect:
            self.form['redirect'] = resolver_redirect
        if not resolver:
            resolver = self.auth.responses.ERROR.response(message=f"Unable to resolve endpoint")
        return resolver

    def ssg_request(self, page, params):
        self.file = page
        self.url = page.data.get('url')
        self.path = page.data.get('url')
        self.slug = page.data.get('slug')
        self.path_params.update(params)

    def redirect(self, url: str):
        """Sets the redirect URL in the request form data"""
        self.form['redirect_to'] = url

    def get_flash_messages(self) -> dict:
        """Pops and returns all flash messages from session"""
        if self.server_request and self.server_request.session:
            session_data = self.server_request.session
            flashes = session_data.pop('__flash__') if session_data.get('__flash__') else {}
            return flashes
        return {}

    def pull_flash(self, key):
        return self.flashes.get(key)

    def add_flash(self, key: str, value: any):
        flash_obj = self.server_request.session.get('__flash__') or {}
        flash_obj[key] = value
        self.server_request.session['__flash__'] = flash_obj

    def from_session(self, session_key: str) -> any:
        """Returns data from the session"""
        return self.server_request.session.get(session_key, None)

    async def process_request_data(self):
        """Get form data and file upload contents from request"""

        from pyonir import Site
        from pyonir.utilities import expand_dotted_keys
        import json

        def secure_upload_filename(filename):
            import re
            # Strip leading and trailing whitespace from the filename
            filename = filename.strip()

            # Replace spaces with underscores
            filename = filename.replace(' ', '_')

            # Remove any remaining unsafe characters using a regular expression
            # Allow only alphanumeric characters, underscores, hyphens, dots, and slashes
            filename = re.sub(r'[^a-zA-Z0-9_.-]', '', filename)

            # Ensure the filename doesn't contain multiple consecutive dots (.) or start with one
            filename = re.sub(r'\.+', '.', filename).lstrip('.')

            # Return the filename as lowercase for consistency
            return filename.lower()

        try:
            try:
                ajson = await self.server_request.json()
                if isinstance(ajson, str): ajson = json.loads(ajson)
                self.form.update(ajson)
            except Exception as ee:
                # multipart/form-data
                form = await self.server_request.form()
                files = []
                for name, content in form.multi_items():
                    if hasattr(content, "filename"):
                        setattr(content, "ext", os.path.splitext(content.filename)[1])
                        self.files.append(content)
                    else:
                        if self.form.get(name): # convert form name into a list
                            currvalue = self.form[name]
                            if isinstance(currvalue, list):
                                currvalue.append(content)
                            else:
                                self.form[name] = [currvalue, content]
                        else:
                            self.form[name] = content
        except Exception as e:
            raise
        self.form = expand_dotted_keys(self.form, return_as_dict=True)

    def derive_status_code(self, is_router_method: bool):
        """Create status code for web request based on a file's availability, status_code property"""
        from pyonir.models.parser import FileStatuses

        code = 404
        if self.file.is_router:
            # If the file is a router method, we assume it is valid
            code = 200
        elif self.file.file_status in (FileStatuses.PROTECTED, FileStatuses.FORBIDDEN):
            self.file.data = {'template': '40x.html', 'content': f'Unauthorized access to this resource.', 'url': self.url, 'slug': self.slug}
            code = 401
        elif self.file.file_status == FileStatuses.PUBLIC or is_router_method:
            code = 200
        self.status_code = code #200 if self.file.file_exists or is_router_method else 404

    def set_context(self) -> None:
        """Gets the routing context from web request"""
        path_str = self.path.replace(self.app.API_ROUTE, '')
        res = None
        for plg in self.app.activated_plugins:
            if not hasattr(plg, 'endpoint'): continue
            if path_str.startswith(plg.endpoint):
                self._app_ctx_ref = plg
                print(f"Request has switched to {plg.name} context")
                break
        return res

    def render_error(self):
        """Data output for an unknown file path for a web request"""
        return {
            "url": self.url,
            "method": self.method,
            "status": self.status_code,
            "res": self.server_response,
            "title": f"{self.path} was not found!",
            "content": f"Perhaps this page once lived but has now been archived or permanently removed from {self.app_ctx_ref.name}."
        }

    def resolve_request_to_file(self, app: BaseApp, path_str: str = None) -> Tuple[BaseApp, Optional['Parsely']]:
        """
        Resolve a request URL to its corresponding resource.

        The function checks plugin-provided paths first, then falls back to the main
        application's file system. If no matching file or virtual route is found,
        a 404 page is returned.
        """
        from pyonir.models.app import BaseApp
        from pyonir.models.page import BasePage, DeserializeFile
        path_str = path_str or self.path
        is_home = path_str == '/'
        ctx_route, ctx_paths = app.request_paths or ('', [])
        ctx_route = ctx_route or ''
        ctx_slug = ctx_route[1:]
        path_slug = path_str[1:]
        app_scope, *path_segments = path_slug.split('/')
        is_api_request = (len(path_segments) and path_segments[0] == app.API_DIRNAME) or path_str.startswith(app.API_ROUTE)

        # Normalize API prefix and path segments
        if is_api_request:
            path_str = path_str.replace(app.API_ROUTE, '')

        request_segments = [
            segment for segment in path_slug.split('/')
            if segment and segment not in (app.API_DIRNAME, ctx_slug)
        ]

        # Skip if no paths or route doesn't match
        if not ctx_paths or (not is_home and not path_str.startswith(ctx_route)):
            return app, None

        # Try resolving to actual file paths
        protected_segment = [s if i > len(request_segments)-1 else f'_{s}' for i,s in enumerate(request_segments)]
        for root_path in ctx_paths:
            if not is_api_request and root_path.endswith(app.API_DIRNAME): continue
            category_index = os.path.join(root_path, *request_segments, 'index.md')
            single_page = os.path.join(root_path, *request_segments) + BaseApp.EXTENSIONS['file']
            single_protected_page = os.path.join(root_path, *protected_segment) + BaseApp.EXTENSIONS['file']

            for candidate in (category_index, single_page, single_protected_page):
                if os.path.exists(candidate):
                    return app, DeserializeFile(candidate, app_ctx=app.app_ctx)

        errorpage = DeserializeFile('404_ERROR')
        errorpage.data = self.render_error()
        return app, errorpage

    @staticmethod
    def process_header(headers):
        nheaders = dict(headers)
        nheaders['accept'] = nheaders.get('accept', TEXT_RES).split(',', 1)[0]
        agent = nheaders.get('user-agent', '')
        nheaders['user-agent'] = agent.split(' ').pop().split('/', 1)[0]
        return nheaders

    @staticmethod
    def get_params(url, as_dict=False):
        import urllib
        from pyonir.utilities import dict_to_class
        args = {params.split('=')[0]: urllib.parse.unquote(params.split("=").pop()) for params in
                url.split('&') if params != ''}
        if args.get('model'): del args['model']
        return args if as_dict else dict_to_class(args, 'query_params')

    def process_resolver(self, request: 'BaseRequest') -> Optional[Union[callable, Any]]:
        """Updates request data a callable method to execute during request."""
        from pyonir import Site
        from pyonir.utilities import get_attr
        resolver_obj = self.file.data.get('@resolvers', {})
        resolver_action = resolver_obj.get(request.method)
        if not resolver_action: return
        resolver_path = resolver_action.get('call')
        del resolver_action['call']
        self.file.data.update(resolver_action)

        app_plugin = list(filter(lambda p: p.name == resolver_path.split('.')[0], Site.activated_plugins))
        app_plugin = app_plugin[0] if len(app_plugin) else Site
        resolver = app_plugin.reload_resolver(resolver_path)

        request.type = get_attr(resolver_action, 'headers.accept') or request.type
        request.form.update(resolver_action)
        if not resolver:
            request.auth.response = request.auth.responses.ERROR.response(message=f"Unable to resolve endpoint")
        return resolver

class BaseServer(Starlette):

    def __init__(self, app: BaseApp):
        self.is_active = False
        self.app = app
        self.ws_routes = []
        self.sse_routes = []
        self.auth_routes = []
        self.endpoints = []
        self.url_map = {}
        self.resolvers = {}
        self.services = {}
        self.request: Optional[BaseRequest] = None
        self.virtual_routes: Optional[OrderedDict] = None

        self.initialize_starlette()


    @property
    def allowed_hosts(self):
        """Returns a list of domains allowed to access the web application"""
        return ['localhost', '*.localhost']

    def run_uvicorn_server(self, uvicorn_options: dict = None):
        """Starts the webserver"""
        import uvicorn, sys
        from pathlib import Path

        # """Uvicorn web server configurations"""
        # Uvicorn’s config only allows one binding method at a time:
        # TCP socket → use host + port (+ optional SSL)
        # Unix domain socket → use uds (+ optional SSL)
        uvicorn_options = uvicorn_options or {}
        if not uvicorn_options:
            if self.app.is_dev:
                uvicorn_options.update({
                    "port": self.app.port,
                    "host": self.app.host
                })
            else:
                uvicorn_options = {'uds': self.app.unix_socket_filepath}

            if self.app.is_secure:
                uvicorn_options["ssl_keyfile"] = self.app.ssl_key_file
                uvicorn_options["ssl_certfile"] = self.app.ssl_cert_file

        # Setup logs
        Path(self.app.logs_dirpath).mkdir(parents=True, exist_ok=True)
        # Initialize routers
        self.init_pyonir_endpoints(self.app)
        print(f"/************** ASGI APP SERVER RUNNING on {'http' if self.app.is_dev else 'sock'} ****************/")
        print(f"\
        \n\t- App env: {'DEV' if self.app.is_dev else 'PROD'}:{self.app.VERSION}\
        \n\t- App name: {self.app.name}\
        \n\t- App domain: {self.app.domain_name}\
        \n\t- App host: {self.app.host}\
        \n\t- App port: {self.app.port}\
        \n\t- App sock: {self.app.unix_socket_filepath}\
        \n\t- App ssl_key: {self.app.ssl_key_file}\
        \n\t- App ssl_cert: {self.app.ssl_cert_file}\
        \n\t- App Server: Uvicorn \
        \n\t- NGINX config: {self.app.nginx_config_filepath} \
        \n\t- System Version: {sys.version_info}")
        print(uvicorn_options)
        self.app.run_plugins(PyonirHooks.AFTER_INIT)
        # self.app.run_hooks(PyonirHooks.AFTER_INIT)
        self.is_active = True
        uvicorn.run(self.app.server, **uvicorn_options)

    def init_app_endpoints(self, router: PyonirRouters):
        for endpoint, routes in router:
            for path, func, methods, *opts in routes:
                self.create_route(func, path=f'{endpoint}{path}', methods=methods, *opts)
                pass

    @staticmethod
    def init_pyonir_endpoints(app: BaseApp):
        from pyonir.tests.backend.demo_controller import pyonir_ws_handler, pyonir_index
        app.server.create_route(pyonir_ws_handler, "/sysws", ws=True)
        app.server.create_route(pyonir_index, "/", methods='*')
        app.server.create_route(pyonir_index, "/{path:path}", methods='*')

    # def create_router(self, endpoints: PyonirRouters):
    #     """A collection (or group) of routes, usually organized by feature or resource, and often mounted under"""
    #     for endpoint, routes in endpoints:
    #         for path, func, methods, *opts in routes:
    #             self.create_route(func, path=f'{endpoint}{path}', methods=methods, *opts)
    #             pass

    @staticmethod
    def create_static_route(assets_dir: str):
        """Creates a route for serving static files"""
        from starlette.staticfiles import StaticFiles
        return StaticFiles(directory=assets_dir)

    def create_route(self, dec_func: Callable,
                       path: str = '',
                       methods=None,
                       auth: bool = None,
                       ws: bool = None,
                       sse: bool = None,
                       static_path: str = None) -> Optional[Callable]:
        """A mapping of an HTTP method and an endpoint to a handler function."""
        import inspect

        is_async = inspect.iscoroutinefunction(dec_func) if dec_func else False
        is_asyncgen = inspect.isasyncgenfunction(dec_func) if dec_func else False
        if methods == '*':
            methods = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE']
        if methods is None:
            methods = ['GET']

        if static_path:
            static_route = self.create_static_route(static_path)
            self.mount(path, static_route)
            self.url_map[path.lstrip('/')
            .replace('/', '_')] = {'path': path, 'dir': static_path, 'exists': os.path.exists(static_path)}
            return None

        route_name = dec_func.__name__
        docs = dec_func.__doc__
        route_path = path.split('/{')[0]
        name = route_name
        endpoint_route = path.split('/', 1)[0]
        # req_models = Site.server.url_map.get(route_name, {}).get('models') or {}

        new_route = {
            "doc": docs,
            "endpoint": endpoint_route,
            "params": {k: v.__name__ for k, v in dec_func.__annotations__.items()},
            "route": path,  # has regex pattern
            "path": route_path,
            "methods": methods,
            "name": name,
            "auth": auth,
            "sse": sse,
            "ws": ws,
            "async": is_async,
            "async_gen": is_asyncgen,
            "func": dec_func
        }
        # Add route path into categories
        self.endpoints.append(f"{endpoint_route}{route_path}")
        self.url_map[name] = new_route
        if sse:
            self.sse_routes.append(f"{endpoint_route}{route_path}")
        if auth:
            self.auth_routes.append(f"{endpoint_route}{route_path}")
        if ws:
            self.ws_routes.append(f"{endpoint_route}{route_path}")
            return self.add_websocket_route(path, dec_func, dec_func.__name__)

        async def dec_wrapper(star_req):
            pyonir_request = BaseRequest(star_req, self.app)
            return await self.build_response(pyonir_request, dec_func)

        dec_wrapper.__name__ = dec_func.__name__
        dec_wrapper.__doc__ = dec_func.__doc__
        self.add_route(path, dec_wrapper, methods=methods)

    def initialize_starlette(self):
        """Setup Starlette web server"""
        super().__init__()
        from starlette_wtf import CSRFProtectMiddleware
        from starlette.middleware.sessions import SessionMiddleware
        from starlette.middleware.trustedhost import TrustedHostMiddleware
        # from starlette.middleware.gzip import GZipMiddleware

        self.add_middleware(TrustedHostMiddleware)
        # star_app.add_middleware(GZipMiddleware, minimum_size=500)
        self.add_middleware(SessionMiddleware,
                                https_only=False,
                                secret_key=self.app.SECRET_SAUCE,
                                session_cookie=self.app.SESSION_KEY,
                                same_site='lax'
                                )
        self.add_middleware(CSRFProtectMiddleware, csrf_secret=self.app.SECRET_SAUCE)

    def get_virtual(self, url: str) -> Union[tuple[str, dict, dict], tuple[None, None, None]]:
        """Performs url pattern matching against virtual routes and returns vitual page data and new path parameter values."""
        for vurl, vdata in self.virtual_routes.items():
            has_match = self.matching_route(url, vurl)
            if has_match:
                return vurl, vdata, has_match
        return None, None, None

    @staticmethod
    async def build_response(pyonir_request: BaseRequest, dec_func: callable):
        import inspect
        from pyonir import Site
        from pyonir.models.auth import Auth
        from pyonir.models.mapper import cls_mapper

        # star_req = pyonir_request.server_request
        Site.server.request = pyonir_request
        default_system_router = dec_func.__name__ == 'pyonir_index'
        # Serve static urls
        if pyonir_request.is_static and default_system_router:
            return Site.server.resolve_static(Site, pyonir_request)

        # Update template globals
        Site.TemplateEnvironment.globals['request'] = pyonir_request

        # Preprocess request form data
        await pyonir_request.process_request_data()
        # Init Auth handler
        pyonir_request.auth = Auth(pyonir_request, Site)
        pyonir_response = BaseRestResponse(status_code=pyonir_request.status_code)

        # Process request context i.e plugin vs app
        pyonir_request.set_context()
        # Process file associated with request
        app_ctx, req_file = pyonir_request.resolve_request_to_file(pyonir_request.app_ctx_ref)
        pyonir_request.file = req_file

        # Preprocess routes or resolver endpoint from file
        app_ctx.apply_virtual_routes(pyonir_request)
        resolver = pyonir_request.process_resolver(pyonir_request)
        # pyonir_request.process_resolver(pyonir_request)

        route_func = dec_func if not callable(resolver) else resolver

        # Check resolver route security
        security = pyonir_request.auth.security.check(pyonir_request.auth)
        if security and not security.accepted:
            return Site.server.serve_redirect(security.redirect_to or '/')

        # Get router endpoint from map
        if callable(route_func):
            is_async = inspect.iscoroutinefunction(route_func)

            default_args = dict(get_type_hints(route_func))
            default_args.update(**pyonir_request.path_params.__dict__)
            default_args.update(**pyonir_request.query_params.__dict__)
            default_args.update(**pyonir_request.form)
            args = cls_mapper(default_args, route_func, from_request=pyonir_request)
            # pyonir_request.router_args = args

            # Resolve route decorator methods
            pyonir_request.server_response = await route_func(**args) if is_async else route_func(**args)

        else:
            pyonir_request.server_response = req_file.resolver or req_file.data

        # Perform redirects
        if pyonir_request.redirect_to:
            return Site.server.serve_redirect(pyonir_request.redirect_to)

        # Derive status code
        pyonir_request.derive_status_code(Site.server.url_map.get(route_func.__name__ if route_func else '') and not default_system_router)

        # Execute plugins hooks initial request
        await Site.run_async_plugins(PyonirHooks.ON_REQUEST, pyonir_request)
        # await Site.run_async_hooks(PyonirHooks.ON_REQUEST, pyonir_request)

        # Finalize response output
        if isinstance(pyonir_request.server_response, BaseRestResponse):
            pyonir_response = pyonir_request.server_response
        else:
            if pyonir_request.type == JSON_RES:
                pyonir_response.set_json(pyonir_request.server_response or pyonir_request.file.data)
            elif pyonir_request.type == TEXT_RES:
                pyonir_response.set_html(pyonir_request.file.output_html(pyonir_request))
            else:
                print(f'{pyonir_request.type} doesnt have a handler')

        # Set headers
        pyonir_response.set_media(pyonir_request.type)
        pyonir_response.set_header('Server', 'Pyonir Web Framework')
        pyonir_response.set_header("Cache-Control", "no-store, no-cache, must-revalidate, max-age=0")
        pyonir_response.set_header("Pragma", "no-cache")
        pyonir_response.set_header("Expires", "0")

        if pyonir_request.is_static:
            return pyonir_request.server_response

        # Generate response
        pyonir_response.set_server_response()
        return pyonir_response.render()

    @staticmethod
    def matching_route(route_path: str, regex_path: str) -> Optional[dict]:
        """Returns path parameters when match is found for virtual routes"""
        from starlette.routing import compile_path
        path_regex, path_format, *args = compile_path(regex_path)
        match = path_regex.match(route_path)# check if request path matches the router regex
        trail_match = match or path_regex.match(route_path+'/')
        if trail_match:
            return trail_match.groupdict()

    @staticmethod
    def resolve_static(app: BaseApp, request: BaseRequest):
        from_frontend = request.path.startswith(app.frontend_assets_route)
        from_public = request.path.startswith(app.public_assets_route)
        base_path = app.frontend_assets_dirpath if from_frontend else app.public_assets_dirpath if from_public else ''
        req_path = request.parts[1:] if len(request.parts) > 1 else request.parts
        path = os.path.join(base_path, *req_path)
        if not os.path.exists(path): print('Pyonir: skipping static: '+ path)
        return BaseServer.serve_static(path)

    @staticmethod
    def serve_static(path: str):
        from starlette.responses import FileResponse, PlainTextResponse
        return FileResponse(path, 200) if os.path.exists(path) else PlainTextResponse(f"{os.path.basename(path)} not found", status_code=404)

    @staticmethod
    def response_renderer(value, media_type):
        from starlette.responses import Response, StreamingResponse
        if media_type == EVENT_RES:
            return StreamingResponse(content=value, media_type=media_type)
        return Response(content=value, media_type=media_type)

    @staticmethod
    def serve_redirect(url: str, code=302):
        from starlette.responses import RedirectResponse
        res = RedirectResponse(url=url.strip(), status_code=code)
        res.headers["Cache-Control"] = "no-store, no-cache, must-revalidate, max-age=0"
        res.headers["Pragma"] = "no-cache"
        res.headers["Expires"] = "0"
        return res

    @staticmethod
    def generate_nginx_conf(app: BaseApp) -> bool:
        """Generates a NGINX conf file based on App configurations"""
        from pyonir.utilities import get_attr, create_file
        nginx_conf = app.TemplateEnvironment.get_template("nginx.jinja.conf") \
            .render(
            app_name=app.name,
            app_name_id=app.name.replace(' ', '_').lower(),
            domain=app.domain_name,
            is_dev=app.is_dev,
            is_secure=app.is_secure,
            ssl_cert_file=app.ssl_cert_file,
            ssl_key_file=app.ssl_key_file,
            site_dirpath=app.app_dirpath,
            site_logs_dirpath=app.logs_dirpath,
            app_socket_filepath=app.unix_socket_filepath,
            app_ignore_logs=f"{app.PUBLIC_ASSETS_DIRNAME}|{app.UPLOADS_DIRNAME}|{app.FRONTEND_ASSETS_DIRNAME}",
            frontend_assets_route=app.frontend_assets_route,
            frontend_assets_dirpath=app.frontend_assets_dirpath,
            public_assets_route=app.public_assets_route,
            public_assets_dirpath=app.public_assets_dirpath,
            site_uploads_route=app.uploads_route,
            site_uploads_dirpath=app.uploads_dirpath,
            site_ssg_dirpath=app.ssg_dirpath,
            custom_nginx_locations=get_attr(app.server, 'nginx_locations')
        )

        return create_file(app.nginx_config_filepath, nginx_conf, False)

@dataclass
class BaseRestResponse:
    """Represents a REST response from the server."""

    status_code: int = 000
    """HTTP status code of the response, e.g., 200 for success, 404 for not found."""

    message: str = ''
    """Response message, typically a string describing the result of the request."""

    data: dict = field(default_factory=dict)
    """Response data, typically a dictionary containing the response payload."""

    _cookies: list = field(default_factory=list)
    _html: str = None
    _stream: any = None
    _media_type: str = None
    _server_response: object = None
    _headers: dict = field(default_factory=dict)

    def to_dict(self) -> dict:
        return {
            'status_code': self.status_code,
            'message': self.message,
            'data': self.data
        }

    def to_json(self) -> str:
        """Converts the response to a JSON serializable dictionary."""
        from pyonir.utilities import json_serial
        import json
        return json.dumps(self.to_dict(), default=json_serial)

    def render(self):
        return self._server_response

    def set_header(self, key, value):
        """Sets header values"""
        self._headers[key] = value

    def set_json(self, value: dict):
        # json = request.file.output_json()
        self.data = value

    def set_html(self, value: str):
        """Sets the html response value"""
        # html = request.file.output_html(request)
        self._html = value

    def set_server_response(self):
        """Renders the starlette web response"""
        from starlette.responses import Response, StreamingResponse

        if self._media_type == EVENT_RES and self._stream:
            return StreamingResponse(content=self._stream, media_type=EVENT_RES)

        content = self._html if self._html else self.to_json()
        media_type = TEXT_RES if self._html else JSON_RES
        self._server_response = Response(content=content, media_type=media_type)
        if self._headers:
            self._server_response.headers.update(self._headers)

    def set_media(self, media_type: str):
        self._media_type = media_type

    def set_cookie(self, cookie: dict):
        """
        :param cookie:
            key="access_token"
            value=jwt_token,
            httponly=True,
            secure=True,
            samesite="strict",
            max_age=3600,
        :return:
        """
        self._cookies.append(cookie)






import os
from dataclasses import dataclass

from pyonir.core import PyonirRequest, PyonirApp

@dataclass
class Menu:
    _orm_options = {'mapper_key': 'menu'}
    url: str
    slug: str = ''
    name: str = ''
    title: str = ''
    group: str = ''
    parent: str = ''
    icon: str = ''
    img: str = ''
    rank: int = 0
    subtitle: str = ''
    dom_class: str = ''
    status: str = ''

    def __post_init__(self):
        if not self.name:
            self.name = self.title
        pass

class Navigation:
    """Assembles a map of navigation menus based on file configurations"""
    name = 'pyonir_navigation'

    def __init__(self, app: PyonirApp):
        self.app = app
        self.menus = {}
        self.active_page = None
        self.build_navigation(app=app)
        # include navigation template example
        # self.register_templates([os.path.join(os.path.dirname(__file__), 'templates')])
        app.TemplateEnvironment.load_template_path(os.path.join(os.path.dirname(__file__), 'templates'))
        self.add_menus_to_environment(app)
        pass

    def after_init(self, data: any, app: PyonirApp):
        self.build_plugins_navigation(app)

    async def on_request(self, request: PyonirRequest, app: PyonirApp):
        """Executes task on web request"""
        refresh_nav = bool(getattr(request.query_params,'rnav', None))
        curr_nav = app.TemplateEnvironment.globals.get('navigation')
        if curr_nav and not refresh_nav: return None
        self.active_page = request.path
        self.build_navigation(app)
        self.add_menus_to_environment(app)

    def add_menus_to_environment(self, app: PyonirApp):
        app.TemplateEnvironment.globals['navigation'] = self.menus.get(app.name)


    def build_plugins_navigation(self, app: PyonirApp):
        if app.activated_plugins:
            for plgn in app.activated_plugins:
                if isinstance(plgn, Navigation):continue
                if not hasattr(plgn, 'pages_dirpath'): continue
                self.build_navigation(plgn)
                pass

    def build_navigation(self, app: PyonirApp):
        # from pyonir.utilities import query_files
        from pyonir.models.database import query_fs
        from collections import defaultdict
        if app is None: return None
        assert hasattr(app, 'pages_dirpath'), "Get menus 'app' parameter does not have a pages dirpath property"
        menus = {}
        submenus = {}
        file_list = query_fs(app.pages_dirpath, app_ctx=app.app_ctx, model=Menu)

        def group_by_menu(items):
            grouped = defaultdict(list)
            for item in items:
                has_menu = item.group or item.parent
                if item.status == 'hidden' or not item.url or (not has_menu): continue
                grouped[item.group].append(item)
            return dict(grouped)  # convert to plain dict if you like
        result = group_by_menu(file_list)
        self.menus[app.name] = result



# from __future__ import annotations
# import os, pytz, re, json
# from datetime import datetime
# from dataclasses import dataclass, field
#
# from .core import PyonirRequest, PyonirCollection
# from .models.database import BasePagination
# from .pyonir_types import ParselyPagination, JSON_RES, AppCtx
# from .utilities import dict_to_class, get_attr, query_files, deserialize_datestr, create_file, get_module, \
#     parse_query_model_to_object, get_file_created
#
# ALLOWED_CONTENT_EXTENSIONS = ('prs', 'md', 'json', 'yaml')
# IGNORE_FILES = ('.vscode', '.vs', '.DS_Store', '__pycache__', '.git', '.', '_', '<', '>', '(', ')', '$', '!', '._')
#
# REG_ILN_LIST = r'([-$@\s*=\w.]+)(\:-)(.*)'
# REG_MAP_LST = r'(^[-$@\s*=\w.]+)(\:[`:`-]?)(.*)'
# REG_METH_ARGS = r"\(([^)]*)\)"
# DICT_DELIM = ": "
# LST_DLM = ":-"
# STR_DLM = ":` "
# ILN_DCT_DLM = ":: "
# BLOCK_DELIM = ":|"
# BLOCK_PREFIX_STR = "==="
# BLOCK_CODE_FENCE = "````"
# LOOKUP_EMBED_PREFIX = '$'
# LOOKUP_FILE_PREFIX = '$file'
# # LOOKUP_CONTENT_PREFIX = '$content'
# LOOKUP_DIR_PREFIX = '$dir'
# FILTER_KEY = '@filter'
# RESOLVER_KEY = '@resolvers'
# DEFAULT_IGNORE_FIELDS = ()
# EmbeddedTypes = dict()
#
# # Image configs
# IMG_FILENAME_DELIM = '|'  # delimits the file name and description
# IMAGE_FORMATS = (
#     'JPEG',  # .jpg, .jpeg
#     'JPG',  # .jpg, .jpeg
#     'PNG',  # .png
#     'GIF',  # .gif
#     'BMP',  # .bmp
#     'TIFF',  # .tiff, .tif
#     'ICO',  # .ico
#     'PPM',  # .ppm
#     'PGM',  # .pgm
#     'PBM',  # .pbm
#     'WebP',  # .webp
#     'DDS',  # .dds
#     'TGA',  # .tga
#     'XBM',  # .xbm
#     'PCX'  # .pcx
# )
#
# class ParselyFileStatus(str):
#     UNKNOWN = 'unknown'
#     """Read only by the system often used for temporary and unknown files"""
#
#     PROTECTED = 'protected'
#     """Requires authentication and authorization. can be READ and WRITE."""
#
#     FORBIDDEN = 'forbidden'
#     """System only access. READ ONLY"""
#
#     PUBLIC = 'public'
#     """Access external and internal with READ and WRITE."""
#
#
# def parse_markdown(content, kwargs):
#     """Parse markdown string using mistletoe with htmlattributesrenderer"""
#     import html, mistletoe
#     # from mistletoe.html_attributes_renderer import HTMLAttributesRenderer
#     if not content: return content
#     res = mistletoe.markdown(content)
#     # res = mistletoe.markdown(content, renderer=HTMLAttributesRenderer)
#     return html.unescape(res)
#
#
# @dataclass
# class Page:
#     """Represents a single page returned from a web request"""
#     _orm_options = {"mapper": {'created_on': 'file_created_on', 'modified_on': 'file_modified_on'}} # avoids merging additional data properties to model
#     url: str
#     is_router: bool = False
#     created_on: datetime = None
#     modified_on: datetime = 'modified_on'
#     tags: list = None
#     date: datetime = None
#     category: str = ''
#     template: str = ''
#     title: str = ''
#     content: str = ''
#     slug: str = ''
#     author: str = 'pyonir'
#     entries: BasePagination = None
#     gallery: dict = None
#     file_name: str = 'file_name'
#     file_path: str = 'file_path'
#     file_dirname: str = 'file_dirname'
#     file_created_on: datetime = 'file_created_on'
#     # contents_relpath: str = ''
#     generate_static_file: callable = None
#     status: str = ParselyFileStatus.PUBLIC
#
#     @property
#     def canonical(self):
#         from pyonir import Site
#         return f"{Site.domain}{self.url}" if Site else None
#
#     def to_json(self) -> dict:
#         """Json serializable repr"""
#         return {k:v for k,v in self.__dict__.items() if k[0]!='_' and k!='app_ctx'}
#
#
#
# @dataclass
# class ParselyMedia:
#     name: str
#     url: str
#     width: int
#     height: int
#     file_size: int
#     thumbnails: dict = field(default_factory=dict)
#     group: str = ''
#     slug: str = ''
#     is_thumb: bool = False
#     captions: str = ''
#     full_url: str = ''
#     file_name: str = ''
#     app_ctx: list = None
#     file_ext: str = ''
#     file_path: str = ''
#     file_dirpath: str = ''
#     file_exists: bool = False
#     file_created_on: datetime = None
#     _sizes: list = field(default_factory=list)
#
#     def open_image(self):
#         from PIL import Image
#         raw_img = Image.open(self.file_path)
#         self.width = get_attr(raw_img, "width", None)
#         self.height = get_attr(raw_img, "height", None)
#         return raw_img
#
#     def to_json(self) -> dict:
#         """Json serializable repr"""
#         from pyonir.utilities import json_serial
#         return {k:json_serial(v) for k,v in self.__dict__.items() if k[0]!='_' and k!='app_ctx'}
#
#     @staticmethod
#     def createImagefolders(folderpath: str):
#         thumbspath = os.path.join(folderpath, 'thumbnails')
#         if not os.path.exists(folderpath):
#             os.makedirs(folderpath)
#         if not os.path.exists(thumbspath):
#             os.makedirs(thumbspath)
#
#     @staticmethod
#     async def save_upload(upload_doc, img_folder_abspath, app_ctx) -> 'ParselyMedia':
#         """Saves base64 file contents into file system"""
#         filename, filedata, rootpath = upload_doc
#         file_name, file_ext = os.path.splitext(filename)
#         ParselyMedia.createImagefolders(img_folder_abspath)
#         new_img_path = os.path.join(img_folder_abspath, file_name + file_ext)
#         file_contents = await filedata.read()
#         with open(new_img_path, 'wb') as f:
#             f.write(file_contents)
#         parselyMedia = ParselyMedia.from_path(str(new_img_path), app_ctx)
#         return parselyMedia
#
#     @classmethod
#     def from_path(cls, file_path, app_ctx):
#         p = Parsely(file_path, app_ctx)
#         return p.map_to_model(cls) if p.file_exists else None
#
#     def resize(self, sizes=None):
#         '''
#         Resize each image and save to the upload path in corresponding image size and paths
#         This happens after full size images are saved to the filesystem
#         '''
#         from PIL import Image
#         from pyonir import Site
#         raw_img = self.open_image()
#         if sizes is None:
#             sizes = [Site.THUMBNAIL_DEFAULT]
#         try:
#             for dimensions in sizes:
#                 width, height = dimensions
#                 self._sizes.append(dimensions)
#                 img = raw_img.resize((width, height), Image.Resampling.BICUBIC)
#                 file_name = f'{self.file_name}--{width}x{height}'
#                 img_dirpath = os.path.dirname(self.file_path)
#                 self.createImagefolders(img_dirpath)
#                 filepath = os.path.join(img_dirpath, Site.UPLOADS_THUMBNAIL_DIRNAME, file_name + '.' + self.file_ext)
#                 if not os.path.exists(filepath): img.save(filepath)
#         except Exception as e:
#             raise
#
#     def generate_thumb(self, width, height) -> str:
#         """Generates an image accordion to width and height parameters and returns url to the new resized image"""
#         self._sizes.append((width, height))
#         if not self.thumbnails.get(f'{width}x{height}'): self.resize([(width, height)])
#         return self.thumbnails.get(f'{width}x{height}')
#
#     def get_all_thumbnails(self) -> dict:
#         """Collects thumbnails for the image"""
#         if self.is_thumb: return None
#         from pyonir import Site
#         if self.group != Site.UPLOADS_DIRNAME: self.group = f'{Site.UPLOADS_DIRNAME}/{self.group}'
#         thumbs_dir = os.path.join(self.file_dirpath, self.group, Site.UPLOADS_THUMBNAIL_DIRNAME)
#         files = query_files(str(thumbs_dir), app_ctx=self.app_ctx, model=ParselyMedia)
#         target_name = self.file_name
#         thumbs = {}
#         # filter files based on name
#         for file in files:
#             if file.file_name[:len(target_name)] != target_name: continue
#             w = file.width
#             h = file.height
#             thumbs[f'{w}x{h}'] = file
#             pass
#         return thumbs
#
# class Parsely:
#     """Parsely is a static file parser"""
#     default_file_attributes = ['file_name','file_path','file_dirpath','file_data_type','file_ctx','file_created_on']
#
#     def __init__(self, abspth: str, app_ctx: AppCtx = None, model: object = None):
#         from pyonir import Site
#         assert abspth is not None, f"Parsely was not provided. {abspth} is not a valid string!"
#         __skip_parsely_deserialization__ = model and hasattr(model, '__skip_parsely_deserialization__')
#         self.schema = model
#         """Model object associated with file."""
#         self._cursor = None
#         self._blob_keys = []
#         self.resolver = None
#         self.is_router = abspth.endswith('.routes.md')
#         self.app_ctx = app_ctx # application context for file
#         self.file_path = str(abspth)
#         self.file_dirpath = os.path.dirname(abspth) # path to files contents directory
#         self.file_virtual_path = ''
#         # file data processing
#         self.file_contents = ''
#         self.file_lines = None
#         self.file_line_count = None
#         self.data = {}
#
#         ctx_url, ctx_name, ctx_dirpath, ctx_staticpath, contents_relpath, file_name, file_ext = self.process_ctx(app_ctx)
#         content_type, *content_subs = os.path.dirname(contents_relpath).split(os.path.sep) # directory at root of contents scope
#         pages_dirname = Site.PAGES_DIRNAME if Site is not None else ''
#         api_dirname = Site.API_DIRNAME if Site is not None else ''
#         is_page = Site is not None and content_type in (pages_dirname, api_dirname) and not self.is_router
#         self.file_contents_dirpath =  ctx_dirpath # contents directory path used when querying refs
#         self.is_page = is_page
#         self.is_home = ctx_url=='' and is_page and contents_relpath == f'{pages_dirname}/index'
#         self.file_ctx = ctx_name # the application context name
#         self.file_dirname = os.path.basename(self.file_dirpath) # nearest parent directory for file
#         self.file_data_type = content_type
#         self.file_name = file_name
#         self.file_ext = file_ext.lstrip('.')
#         self.status = self.file_status
#
#         surl = re.sub(fr'\b{pages_dirname}/\b|\bindex\b', '', contents_relpath) if is_page else contents_relpath
#         slug = f'{ctx_url}/{surl}'.lstrip('/').rstrip('/').lower()
#         url = '/' if self.is_home else '/' + slug
#         # page attributes
#         if is_page:
#             self.data['url']  = url
#             self.data['slug'] = slug
#             self.data['tags'] = content_subs
#         self.file_ssg_api_dirpath = os.path.join(ctx_staticpath, api_dirname, slug)
#         self.file_ssg_html_dirpath = os.path.join(ctx_staticpath, slug)
#         if __skip_parsely_deserialization__: return
#         self.deserializer()
#         self.apply_filters()
#
#
#     @property
#     def file_status(self) -> str:  # String
#         if not self.file_exists: return ParselyFileStatus.UNKNOWN
#         return ParselyFileStatus.PROTECTED if self.file_name.startswith('_') else \
#             ParselyFileStatus.FORBIDDEN if self.file_name.startswith('.') else ParselyFileStatus.PUBLIC
#
#     @property
#     def file_created_on(self):  # Datetime
#         return get_file_created(self.file_path) if self.file_exists else None
#
#     @property
#     def file_modified_on(self):  # Datetime
#         return datetime.fromtimestamp(os.path.getmtime(self.file_path), tz=pytz.UTC) if self.file_exists else None
#
#     @property
#     def file_exists(self):
#         return os.path.exists(self.file_path) if self.file_path else None
#
#     @property
#     def template(self) -> str:
#         if not self.file_exists or not isinstance(self.data, dict) and not self.is_router: return "40x.html"
#         return self.data.get('template', "pages.html")
#
#     @property
#     def date(self):
#         """Returns manual published date but defaults to file_created_on date"""
#         file_date = self.data.get('date', self.file_created_on) if isinstance(self.data, dict) else self.file_created_on
#         if isinstance(file_date, str):
#             file_date = deserialize_datestr(file_date)
#         return file_date
#
#     def process_ctx(self, app_ctx = None):
#         ctx_url = ''
#         if not app_ctx:
#             ctx_pages_dirpath = os.path.dirname(self.file_path)
#             ctx_dir = os.path.basename(ctx_pages_dirpath)
#             ctx_staticpath = os.path.join(ctx_pages_dirpath, 'ssg')
#         else:
#             ctx_dir, ctx_url, ctx_pages_dirpath, ctx_staticpath = app_ctx
#         if not os.path.exists(ctx_pages_dirpath): ctx_pages_dirpath = os.path.dirname(ctx_pages_dirpath)
#         _, _, content_dirpath = self.file_path.partition(ctx_pages_dirpath)
#         file_name, file_ext = os.path.splitext(os.path.basename(self.file_path))
#         content_dirpath = content_dirpath.lstrip(os.path.sep).replace(file_ext,'')
#         return ctx_url or '', ctx_dir, ctx_pages_dirpath, ctx_staticpath, content_dirpath, file_name, file_ext
#
#     def set_taxonomy(self) -> list[str]:
#         if not self.file_exists: return None
#         # cat_list = os.path.dirname(self.file_relpath).lstrip('/').split('/')
#         self.data['tags'] = self.data.get('tags')
#         self.data['category'] = self.data.get('category')
#
#     def set_media(self):
#         """Sets data property for media files. (jpeg, png, pdf, etc)"""
#         if self.file_ext.endswith(ALLOWED_CONTENT_EXTENSIONS): return
#         from PIL import Image
#         from pyonir import Site
#         group = self.file_dirpath.lstrip('/').split(Site.uploads_route, 1).pop()
#         group_name = group.lstrip('/').split(os.path.sep)[0]
#         image_name, *image_captions = self.file_name.replace('.' + self.file_ext, '').split(IMG_FILENAME_DELIM)
#         formatted_name = re.sub(r'[^a-zA-Z0-9]+', ' ', image_name).title()
#         formated_caption = "".join(image_captions or formatted_name).title()
#         _slug = f"{Site.UPLOADS_DIRNAME}/{group_name+'/' if group else ''}{image_name}.{self.file_ext}"
#         _url = f"{_slug}"
#         is_thumb = Site.UPLOADS_THUMBNAIL_DIRNAME in self.file_dirname
#         full_url = _slug.replace(Site.UPLOADS_THUMBNAIL_DIRNAME + '/', '').split('--')[0] + f".{self.file_ext}" if is_thumb else _url
#         raw_img = Image.open(self.file_path)
#         width = get_attr(raw_img, "width", None)
#         height = get_attr(raw_img, "height", None)
#
#
#         self.data = {
#             'file_size': os.path.getsize(self.file_path),
#             'url': _url,
#             'slug': _slug,
#             'name': formatted_name,
#             'full_url': full_url,
#             'group': group_name,
#             'is_thumb': is_thumb,
#             'captions': formated_caption,
#             'width':width, 'height': height,
#             # 'raw_img': raw_img,
#         }
#         pass
#
#     def map_to_model(self, model, refresh = False):
#         """Maps the parsely object into a model class provided or the self.schema reference """
#         from pyonir.models.mapper import cls_mapper
#         if model is None:
#             model = self.schema
#         if model and refresh:
#             import sys, importlib
#             model_name = model.__name__ if hasattr(model, '__name__') else type(model).__name__
#             module_path = model.__module__ if hasattr(model, '__module__') else None
#             module = sys.modules.get(module_path) if module_path else None
#             if module:
#                 print(f'pyonir is reloading the model:{module_path}.{model_name}')
#                 sys.modules[module_path] = importlib.reload(module)
#                 new_model = getattr(sys.modules[module_path], model_name)
#                 model = new_model
#
#         if not model:
#             file_props = {k: v for k,v in self.__dict__.items() if k in self.default_file_attributes}
#             return dict_to_class({**self.data, **file_props, '@model': 'GenericQueryModel'}, self.file_data_type, True)
#         try:
#             res = cls_mapper(self, model)
#             return res
#         except Exception as e:
#             raise
#
#     def to_json(self) -> dict:
#         """Json serializable repr"""
#         return self.data
#
#     def apply_template(self, prop_names: list = None, context: dict = None):
#         """Render python format strings for data property values"""
#         from pyonir import Site
#         context = context or self.data
#         prop_names = context.get('@pyformatter', [])
#         # prop_names = prop_names or self.data.get(FILTER_KEY,{}).get("pyformat", self.data.get("@pyformatter")) or context.get('@pyformatter', [])
#         for prop in prop_names:
#             data_value = context.get(prop)
#             data_value = Site.parse_pyformat(data_value, self.data)
#             update_nested(prop, data_src=self.data, data_update=data_value)
#
#
#     def apply_filters(self):
#         """Applies filter methods to data attributes"""
#         if not bool(self.data): return
#         filters = self.data.get(FILTER_KEY)
#         if not filters: return
#         # from pyonir import Site
#         for filtr, datakeys in filters.items():
#             ifiltr = self.app_filter(filtr)
#             if not ifiltr: continue
#             for key in datakeys:
#                 mod_val = ifiltr(get_attr(self.data, key), {"page": self.data})
#                 update_nested(key, self.data, data_update=mod_val)
#         del self.data[FILTER_KEY]
#
    # @staticmethod
    # def load_resolver(relative_module_path: str, base_path: str = '', from_system: bool = False):
    #     from pyonir.utilities import import_module
    #     if '.' not in relative_module_path: return None
    #     pkg = relative_module_path.split('.')
    #     if from_system: base_path = os.path.dirname(base_path)
    #     meth_name = pkg.pop()
    #     pkg_path = ".".join(pkg)
    #     module_base = pkg[:-1]
    #     module_name = pkg[-1]
    #     _pkg_dpath = os.path.join(base_path, *module_base) + '.py' # is a /path/to/module
    #     _module_dpath = os.path.join(base_path, *module_base, module_name+'.py') # is a /path/to/module.py
    #     _module_pkg_dpath = os.path.join(base_path, *pkg, '__init__.py') # is a /path/to/module/__init__.py
    #     if os.path.exists(_pkg_dpath):
    #         pkg_path = ".".join(module_base)
    #         meth_name = f"{module_name}.{meth_name}"
    #     elif os.path.exists(_module_dpath):
    #         meth_name = f"{module_name}.{meth_name}"
    #     elif os.path.exists(_module_pkg_dpath):
    #         pass
    #     else:
    #         return None
    #     module_callable = import_module(pkg_path, callable_name=meth_name)
    #     return module_callable
#
#     async def _access_module_from_request(self, resolver_path: str) -> tuple:
#         from pyonir import Site
#         if resolver_path.startswith(LOOKUP_DIR_PREFIX):
#             return None, self.process_value_type(resolver_path)
#         app_plugin = list(filter(lambda p: p.name == resolver_path.split('.')[0], Site.activated_plugins))
#         app_plugin = app_plugin[0] if len(app_plugin) else Site
#         resolver = app_plugin.reload_resolver(resolver_path)
#
#         return None, resolver
#
#     async def process_resolver(self, request: PyonirRequest):
#         """Resolves dynamic data from external methods"""
#         resolver_obj = self.data.get(RESOLVER_KEY, {})
#         resolver_action = resolver_obj.get(request.method)
#         if resolver_obj and not resolver_action:
#             self.status = ParselyFileStatus.FORBIDDEN
#             request.type = JSON_RES
#         if not resolver_action or self.resolver is not None: return
#         resolver_path = resolver_action.get('call')
#         resolver_args = resolver_action.get('args')
#         resolver_redirect = resolver_action.get('redirect')
#
#         if not resolver_path: return
#         self.data = resolver_action
#         module, resolver = await self._access_module_from_request(resolver_path)
#
#         request.type = resolver_action.get('headers', {}).get('accept', request.type)
#         if resolver and resolver_args:
#             request.form.update(resolver_args)
#         if resolver and resolver_redirect:
#             request.form['redirect'] = resolver_redirect
#         if not resolver:
#             resolver = request.auth.responses.ERROR.response(message=f"Unable to resolve endpoint")
#         self.resolver = resolver
#
#     def deserializer(self):
#         """Deserialize file line strings into map object"""
#         if not self.file_exists:
#             return
#         if self.file_ext == 'md' or self.file_contents:
#             self.process_setup()
#             if self.file_line_count > 0:
#                 self.process_line(0, output_data=self.data)
#         elif self.file_ext == 'json':
#             self.data = self.open_file(self.file_path, rtn_as='json') or {}
#         elif self.file_ext.upper() in IMAGE_FORMATS:
#             self.set_media()
#         return True
#
#     def process_setup(self):
#         lines = self.file_contents or ''
#         if self.file_exists and self.file_ext == 'md':
#             lines = self.open_file(self.file_path)
#         self.file_lines = lines.strip().split("\n")
#         self.file_contents = "\n".join(self.file_lines)
#         self.file_line_count = len(self.file_lines)
#
#     def process_line(self, cursor, output_data: any = None, is_blob=None, stop_str: str = '') -> tuple:
#         """Deserializes string value"""
#
#         def count_tabs(str_value: str, tab_width: int = 4):
#             """Returns number of tabs for provided string"""
#             try:
#                 return round(len(re.match(r'^\s+', str_value.replace('\n', '')).group()) / tab_width)
#             except Exception as e:
#                 return 0
#
#         def process_iln_frag(ln_frag, val_type=None):
#             """processing inline values for nested objects"""
#
#             def get_pairs(ln_frag):
#                 """partition key value pairs"""
#                 # faster than regex! 0.01ms vs 0.45ms
#                 try:
#                     methArgs = ''
#                     if ln_frag.endswith(DICT_DELIM.strip()):
#                         return (ln_frag[:-1], DICT_DELIM, "") + (methArgs,)
#                     iln_delim = [x for x in (
#                         (ln_frag.find(STR_DLM), STR_DLM),
#                         (ln_frag.find(LST_DLM), LST_DLM),
#                         (ln_frag.find(DICT_DELIM), DICT_DELIM),
#                     ) if x[0] != -1]
#                     return ln_frag.partition(iln_delim[0][1]) + (methArgs,)
#                 except Exception as e:
#                     # print(f"`{ln_frag.strip()}` >>> {e}")
#                     return (None, None, ln_frag.strip(), None)
#
#             keystr, delim, valuestr, methargs = get_pairs(ln_frag)
#
#             parsed_key = keystr.strip() if keystr and keystr.strip() != '' else None
#             val_type = get_container_type(delim) if val_type is None else val_type
#             parsed_val = valuestr.strip()
#             force_scalr = delim and delim.endswith('`') or parsed_val.startswith(LOOKUP_DIR_PREFIX)
#             is_inline_expression = bool(parsed_key and parsed_val) and not force_scalr
#             if is_inline_expression:
#                 has_dotpath = "." in parsed_key
#                 if has_dotpath or (isinstance(val_type, list) and (", " in parsed_val)):  # inline list
#                     data_container = [] if delim is None else val_type #get_container_type(delim)
#                     for x in parsed_val.split(', '):
#                         pk, vtype, pv, pmethArgs = process_iln_frag(x)
#                         if vtype != '' and pk:
#                             _, pv = update_nested(pk, vtype, pv)
#                         update_nested(None, data_container, pv)
#                     parsed_val = data_container or pv
#                 elif isinstance(val_type, list):
#                     parsed_val = [parsed_val]
#                     # val_type.append(parsed_val)
#
#             # skip_line = hasattr(self.schema, parsed_key) if parsed_key and self.schema else None
#             parsed_val = self.process_value_type(parsed_val) #if not skip_line else parsed_val
#
#             return parsed_key, val_type, parsed_val, methargs
#
#         def get_container_type(delim):
#             if LST_DLM == delim:
#                 return list()
#             elif DICT_DELIM == delim:
#                 return dict()
#             else:
#                 return str()
#
#
#         def stop_loop_block(cur, curtabs, is_blob=None, stop_str=None):
#             if cur == self.file_line_count: return True
#             in_limit = cur + 1 < self.file_line_count
#             stop_comm_blok = self.file_lines[cur].strip().endswith(stop_str) if in_limit and stop_str else None
#             nxt_curs_is_blok = in_limit and self.file_lines[cur + 1].startswith(BLOCK_PREFIX_STR)
#             nxt_curs_is_blokfence = in_limit and self.file_lines[cur + 1].strip().startswith(BLOCK_CODE_FENCE)
#             nxt_curs_is_blokdelim = in_limit and self.file_lines[cur + 1].strip().endswith(BLOCK_DELIM)
#             nxt_curs_tabs = count_tabs(self.file_lines[cur + 1]) if (in_limit and not is_blob) else -1
#             res = True if stop_comm_blok or nxt_curs_is_blokfence or nxt_curs_is_blok or nxt_curs_is_blokdelim or\
#                 (nxt_curs_tabs < curtabs and not is_blob) else False
#             return res
#             # return '__STOPLOOKAHEAD__' if stop_comm_blok or nxt_curs_is_blok or nxt_curs_is_blokdelim or (
#             #         nxt_curs_tabs < curtabs and not is_blob) else None
#         stop = False
#         stop_iter = False
#         while cursor < self.file_line_count:
#             self._cursor = cursor
#             if stop: break
#             ln_frag = self.file_lines[cursor]
#             is_multi_ln_comment = ln_frag.strip().startswith('{#')
#             is_block_code = ln_frag.strip().startswith(BLOCK_CODE_FENCE)
#             is_end_block_code = ln_frag.strip() == BLOCK_CODE_FENCE
#             is_ln_comment = not is_blob and ln_frag.strip().startswith('#') or not is_blob and ln_frag.strip() == ''
#             comment = is_multi_ln_comment or is_ln_comment
#
#             if comment or is_end_block_code:
#                 if is_multi_ln_comment or stop_str:
#                     cursor, ln_val = self.process_line(cursor + 1, '', stop_str='#}')
#             else:
#                 tabs = count_tabs(ln_frag)
#                 stop_iter = tabs > 0 and not is_ln_comment or is_blob or stop_str
#                 try:
#                     if is_blob:
#                         output_data += ln_frag + "\n"
#                     elif not comment and not stop_str:
#                         inlimts = cursor + 1 < self.file_line_count
#                         is_block = ln_frag.startswith(BLOCK_PREFIX_STR) or ln_frag.endswith("|") or is_block_code
#                         # TODO: is_parent should be less restrictive on tabs vs spaces.
#                         is_parent = True if is_block else count_tabs(
#                             self.file_lines[cursor + 1]) > tabs if inlimts else False
#                         parsed_key, val_type, parsed_val, methArgs = process_iln_frag(ln_frag)
#                         # if methArgs:
#                         #     output_data['@args'] = [arg.replace(' ', '').split(':') for arg in methArgs.split(',')]
#                         if is_parent or is_block:
#                             parsed_key = parsed_val if not parsed_key else parsed_key
#                             parsed_key = "content" if parsed_key == BLOCK_PREFIX_STR else \
#                                 (parsed_key.replace(BLOCK_PREFIX_STR, "")
#                                  .replace(BLOCK_DELIM,'')
#                                  .replace(BLOCK_CODE_FENCE,'').strip())
#                             if is_block_code:
#                                 fence_key, *overide_keyname = parsed_key.split(' ', 1)
#                                 parsed_key = overide_keyname[0] if overide_keyname else fence_key
#                                 pass
#                             cursor, parsed_val = self.process_line(cursor + 1, output_data=val_type, is_blob=isinstance(val_type, str))
#                             if isinstance(parsed_val, list) and '-' in parsed_val:  # consolidate list of maps
#                                 parsed_val = self.post_process_blocklist(parsed_val)
#
#                         # Store objects with $ prefix
#                         if parsed_key and parsed_key.startswith('$'):
#                             EmbeddedTypes[parsed_key] = parsed_val
#                         else:
#                             # Extend objects that inheirit from other files during post processing
#                             if parsed_key == '@extends':
#                                 if not isinstance(parsed_val, dict):
#                                     print(f'{self.file_path}')
#                                 output_data.update(parsed_val)
#                                 output_data['@extends'] = ln_frag.split(':').pop().strip()
#                             else:
#                                 _, output_data = update_nested(parsed_key, output_data, data_merge=parsed_val)
#                 except Exception as e:
#                     # raise Exception(f"{self.file_name}: {str(e)}")
#                     raise
#
#             stop = stop_loop_block(cursor, tabs, is_blob, stop_str=stop_str) if stop_iter else None
#             if not stop: cursor += 1
#
#         return cursor, output_data
#
#     def process_value_type(self, valuestr: str):
#         """Deserialize string value to appropriate object type"""
#         if not isinstance(valuestr, str): return valuestr
#
#         def is_num(valstr):
#             valstr = valstr.strip().replace(',', '')
#             if valstr.isdigit():
#                 return int(valstr)
#             try:
#                 return float(valstr)
#             except ValueError:
#                 return 'NAN'
#
#         valuestr = valuestr.strip()
#         if EmbeddedTypes.get(valuestr):
#             return EmbeddedTypes.get(valuestr)
#
#         isnum = is_num(valuestr)
#         if isnum != 'NAN':
#             return isnum
#         if valuestr.strip().lower() == "false":
#             return False
#         elif valuestr.strip().lower() == "true":
#             return True
#         elif valuestr.strip().startswith('$'):
#
#             def parse_ref_to_files(filepath, as_dir=0):
#                 # use proper app context for path reference outside of scope is always the root level
#                 # Ref parameters with model will return a generic model to represent the data value
#                 base_path = os.path.dirname(self.app_ctx[2])
#                 generic_query_model = self.load_resolver(generic_model_properties, base_path=base_path) if generic_model_properties else None
#                 generic_query_model = generic_query_model or parse_query_model_to_object(generic_model_properties)
#
#                 if as_dir:
#                     collection = PyonirCollection.query(filepath,
#                                         app_ctx=self.app_ctx,
#                                         force_all=return_all_files,
#                                         model=generic_query_model,
#                                         exclude_names=(self.file_name + '.' + self.file_ext, 'index.md')
#                                                       )
#                     data = collection.paginated_collection(query_params)
#                 else:
#                     rtn_key = has_attr_path or 'data'
#                     p = Parsely(filepath, self.app_ctx)
#                     data = get_attr(p, rtn_key) or p
#                 # EmbeddedTypes[filepath] = d
#                 return data
#
#             cvaluestr = valuestr.strip()
#             valuestr = valuestr.strip()
#             has_file_ref = valuestr.startswith(LOOKUP_FILE_PREFIX)
#             has_dir_ref = valuestr.startswith(LOOKUP_DIR_PREFIX)
#             if '{{' in valuestr:
#                 jinja = self.app_filter('jinja')
#                 valuestr = jinja(valuestr, self.data)
#             if valuestr.startswith('$') and '{' in valuestr:
#                 pyformat = self.app_filter('pyformat')
#                 valuestr = pyformat(valuestr[1:] if not has_dir_ref else valuestr, self.__dict__) if pyformat else valuestr
#             if has_dir_ref:
#                 query_params = valuestr.split("?").pop() if "?" in valuestr else False
#                 has_attr_path = valuestr.split("#")[-1] if "#" in valuestr else ''
#                 valuestr = valuestr.replace(f"{LOOKUP_DIR_PREFIX}/", "") \
#                     .replace(f"?{query_params}", "") \
#                     .replace(f'#{has_attr_path}', '')
#                 query_params = dict(map(lambda x: x.split("="), query_params.split('&')) if query_params else '')
#                 use_rel = valuestr.startswith('./')
#                 return_all_files = valuestr.endswith('/*')
#                 generic_model_properties = query_params.get('model')
#                 valuestr = valuestr.replace('../', '').replace('/*', '')
#                 dir_root = self.file_contents_dirpath if not use_rel else self.file_dirpath
#                 lookup_fpath = os.path.join(dir_root, *valuestr.split("/"))
#                 if not os.path.exists(lookup_fpath):
#                     print({
#                         'ISSUE': f'FileNotFound while processing {cvaluestr}',
#                         'SOLUTION': f'Make sure the `{lookup_fpath}` file exists. Note that only valid md and json files can be processed.'
#                     })
#                     return cvaluestr if self.is_router else None
#                 return EmbeddedTypes.get(lookup_fpath, parse_ref_to_files(lookup_fpath, os.path.isdir(lookup_fpath)))
#
#         return valuestr.lstrip('$')
#
#     @staticmethod
#     def serializer(json_map: any, namespace: list = [], inline_mode: bool = False, filter_params=None) -> str:
#         """Converts json string into parsely string"""
#
#         if filter_params is None:
#             filter_params = {}
#         mode = 'INLINE' if inline_mode else 'NESTED'
#         lines = []
#         multi_line_keys = []
#         is_block_str = False
#
#         def pair_map(key, val, tabs):
#             is_multiline = isinstance(val, str) and len(val.split("\n")) > 2
#             if is_multiline or key in filter_params.get('_blob_keys', []):
#                 multi_line_keys.append((f"==={key.replace('content', '')}{filter_params.get(key, '')}", val.strip()))
#                 return
#             if mode == 'INLINE':
#                 ns = ".".join(namespace)
#                 value = f"{ns}.{key}: {val}" if bool(namespace) else f"{key}: {val.strip()}"
#                 lines.append(value)
#             else:
#                 # if multiline:
#                 #     lines.append(f"=== {key}\n{val}")
#                 if key:
#                     lines.append(f"{tabs}{key}: {val}")
#                 else:
#                     lines.append(f"{tabs}{val}")
#
#         if isinstance(json_map, (str, bool, int, float)):
#             tabs = '    ' * len(namespace)
#             return f"{tabs}{json_map}"
#
#         for k, val in json_map.items():
#             tab_count = len(namespace) if namespace is not None else 0
#             tabs = '    ' * tab_count
#             # block_prefx = BLOCK_PREFIX_STR + ' ' if tab_count == 0 else ''
#             # print('\t'*len(namespace), k)
#             if isinstance(val, (str, int, bool, float)):
#                 pair_map(k, val, tabs)
#
#             elif isinstance(val, (dict, list)):
#                 delim = ':' if isinstance(val, dict) else ':-'
#                 if len(namespace) > 0:
#                     namespace = namespace + [k]
#                 else:
#                     namespace = [k]
#
#                 if mode == 'INLINE' and isinstance(val, list):
#                     ns = ".".join(namespace)
#                     lines.append(f"{ns}{delim}")
#                 elif mode == 'NESTED':
#                     lines.append(f"{tabs}{k}{delim}")
#
#                 if isinstance(val, dict):
#                     nested_value = Parsely.serializer(json_map=val, namespace=namespace, inline_mode=inline_mode)
#                     lines.append(f"{nested_value}")
#                 else:
#                     maxl = len(val) - 1
#                     has_scalar = any([isinstance(it, (str, int, float, bool)) for it in val])
#                     for i, item in enumerate(val):
#                         list_value = Parsely.serializer(json_map=item, namespace=namespace, inline_mode=False)
#                         lines.append(f"{list_value}")
#                         if i < maxl and not has_scalar:
#                             lines.append(f"    -")
#                 namespace.pop()
#
#         if multi_line_keys:
#             [lines.append(f"{mlk}\n{mlv}") for mlk, mlv in multi_line_keys]
#         return "\n".join(lines)
#
#     @staticmethod
#     def open_file(file_path: str, rtn_as: str = 'string'):
#         """Reads target file on file system"""
#         if not os.path.exists(file_path): return None
#         with open(file_path, 'r', encoding='utf-8') as target_file:
#             try:
#                 if rtn_as == "list":
#                     return target_file.readlines()
#                 elif rtn_as == "json":
#                     return json.load(target_file)
#                 else:
#                     return target_file.read()
#             except Exception as e:
#                 return {"error": __file__, "message": str(e)} if rtn_as == "json" else []
#
#     @staticmethod
#     def post_process_blocklist(blocklist: list):
#         if not isinstance(blocklist, list): return blocklist
#
#         def merge(src, trg):
#             ns = []
#             for k in src.keys():
#                 tv = trg.get(k)
#                 if tv:
#                     ns.append(k)
#                     trg = trg.get(k)
#
#             update_nested(ns, src, trg)
#             return src
#
#         _temp_list_obj = {}  # used for blocks that have `-` separated maps
#         results = []
#         max_count = len(blocklist)
#         for i, hashitem in enumerate(blocklist):
#             if isinstance(hashitem, dict):
#                 _temp_list_obj = merge(_temp_list_obj, hashitem)
#                 if i + 1 == max_count:
#                     results.append(dict(_temp_list_obj))
#                     break
#             else:
#                 results.append(dict(_temp_list_obj))
#                 _temp_list_obj.clear()
#         blocklist = results
#         return blocklist
#
#     @classmethod
#     def from_input(cls, input_src: dict, app_ctx: tuple):
#         """Creates Parsely object setting the data from input src"""
#         if not input_src: return None
#         res = cls('', app_ctx)
#         res.data = input_src
#         return res
#
#     @classmethod
#     def create_file(cls, file_path: str, contents: dict, app_ctx: tuple) -> 'Parsely':
#         """Creates new file on filesystem"""
#         dir_path = os.path.dirname(file_path)
#         is_json = file_path.endswith('json')
#         if dir_path and not os.path.exists(dir_path):
#             os.makedirs(dir_path)
#         new_file = create_file(
#             file_path,
#             contents,
#             is_json=is_json,
#             mode='w+')
#         new_parsely_file = cls(file_path, app_ctx) if new_file else None
#         return new_parsely_file
#
#     def throw_error(self, message: dict):
#         msg = {
#             'ERROR': f'{self.file_path} found an error on line {self._cursor}',
#             'LINE': f'{self.file_lines[self._cursor]}', **message}
#         return msg
#
#     def app_filter(self, filter_name: str):
#         from pyonir import Site
#         return Site.Parsely_Filters.get(filter_name) if Site else {}
#
#     def refresh_data(self):
#         """Parses file and update data values"""
#         self.data = {}
#         self._blob_keys.clear()
#         self.deserializer()
#         self.apply_filters()
#
#     def prev_next(self):
#         if self.file_dirname != 'pages' or self.is_home:
#             return None
#         return PyonirCollection.prev_next(self)
#
#     def output_json(self, data_value: any = None, as_str=True) -> dict:
#         """Outputs a json string"""
#         from .utilities import json_serial
#         data = data_value or self
#         if not as_str: return data
#         return json.dumps(data, default=json_serial)
#
#     def output_html(self, req: PyonirRequest) -> str:
#         """Renders and html output"""
#         from pyonir import Site
#         refresh_model = get_attr(req, 'query_params.rmodel')
#         page = self.map_to_model(Page, refresh=refresh_model)
#         Site.apply_globals({'prevNext': self.prev_next, 'page': page})
#         html = Site.TemplateEnvironment.get_template(page.template).render()
#         Site.TemplateEnvironment.block_pull_cache.clear()
#         return html
#
#     def save(self, file_path=None, contents=None, dir_path=None) -> bool:
#         """Saves data into file_path"""
#         file_path = file_path or self.file_path
#         is_json = file_path.endswith('json')
#         if dir_path and not os.path.exists(dir_path):
#             os.makedirs(dir_path)
#         return create_file(
#             file_path,
#             contents or self.data,
#             is_json=is_json,
#             mode='w+')
#
#     def generate_static_file(self, page_request=None, rtn_results=False):
#         """Generate target file as html or json. Takes html or json content to save"""
#         # from pyonir import Site
#         # from htmlmin import minify
#
#         count = 0
#         html_data = None
#         json_data = None
#
#         def render_save():
#             # -- Render Content --
#             html_data = self.output_html(page_request)
#             json_data = self.output_json(as_str=False)
#             # -- Save contents --
#             self.save(path_to_static_html, html_data, self.file_ssg_html_dirpath)
#             self.save(path_to_static_api, json_data, self.file_ssg_api_dirpath)
#             return 2
#
#         # -- Get static paths --
#         path_to_static_api = os.path.join(self.file_ssg_api_dirpath, "index.json")
#         path_to_static_html = os.path.join(self.file_ssg_html_dirpath, "index.html")
#
#         count += render_save()
#
#         if page_request:
#             for pgnum in range(1, page_request.paginate):
#                 path_to_static_html = os.path.join(self.file_ssg_html_dirpath, str(pgnum + 1), "index.html")
#                 path_to_static_api = os.path.join(self.file_ssg_api_dirpath, str(pgnum + 1), "index.json")
#                 page_request.query_params['pg'] = pgnum + 1
#                 count += render_save()
#
#         # -- Return contents without saving --
#         if rtn_results:
#             return html_data, json_data
#
#         return count
#
#
# def update_nested(attr_path, data_src: dict, data_merge=None, data_update=None, find=None) -> tuple[bool, dict]:
#     """
#     Finds or updates target value based on an attribute path.
#
#     Args:
#         attr_path (list): Attribute path as list or dot-separated string.
#         data_src (dict): Source data to search or update.
#         data_merge (Any, optional): Value to merge.
#         data_update (Any, optional): Value to replace at path.
#         find (bool, optional): If True, only retrieve the value.
#
#     Returns:
#         tuple[bool, Any]: (completed, updated data or found value)
#     """
#
#     def update_value(target, val):
#         """Mutates target with val depending on type compatibility."""
#         if isinstance(target, list):
#             if isinstance(val, list):
#                 target.extend(val)
#             else:
#                 target.append(val)
#         elif isinstance(target, dict) and isinstance(val, dict):
#             target.update(val)
#         elif isinstance(target, str) and isinstance(val, str):
#             return val
#         return target
#
#     # Normalize attribute path
#     if isinstance(attr_path, str):
#         attr_path = attr_path.strip().split('.')
#     if not attr_path:
#         return True, update_value(data_src, data_merge)
#
#     completed = len(attr_path) == 1
#
#     # Handle list source at top-level
#     if isinstance(data_src, list):
#         _, merged_val = update_nested(attr_path, {}, data_merge)
#         return update_nested(None, data_src, merged_val)
#
#     # Navigate deeper if not at last key
#     if not completed:
#         current_data = {}
#         for i, key in enumerate(attr_path):
#             if find:
#                 current_data = (data_src.get(key) if not current_data else current_data.get(key))
#             else:
#                 completed, current_data = update_nested(
#                     attr_path[i + 1:],
#                     data_src.get(key, current_data),
#                     find=find,
#                     data_merge=data_merge,
#                     data_update=data_update
#                 )
#                 update_value(data_src, {key: current_data})
#                 if completed:
#                     break
#     else:
#         # Last key operations
#         key = attr_path[-1].strip()
#
#         if find:
#             return True, data_src.get(key)
#
#         if data_update is not None:
#             return completed, update_value(data_src, {key: data_update})
#
#         # If key not in dict, wrap merge value in a dict
#         if isinstance(data_src, dict) and data_src.get(key) is None:
#             data_merge = {key: data_merge}
#
#         if isinstance(data_merge, (str, int, float, bool)):
#             data_src[key] = data_merge
#         elif isinstance(data_src, dict):
#             update_value(data_src.get(key, data_src), data_merge)
#         else:
#             update_value(data_src, data_merge)
#
#     return completed, (data_src if not find else current_data)



import base64
import re
from datetime import datetime
import os, pytz
from pathlib import Path
from typing import Any, Union, Optional

from pyonir.models.database import BasePagination
from pyonir.models.parser import DeserializeFile
from pyonir.utilities import get_attr

IMG_FILENAME_DELIM = '::'  # delimits the file name and description

class PageStatus(str):
    UNKNOWN = 'unknown'
    """Read only by the system often used for temporary and unknown files"""

    PROTECTED = 'protected'
    """Requires authentication and authorization. can be READ and WRITE."""

    FORBIDDEN = 'forbidden'
    """System only access. READ ONLY"""

    PUBLIC = 'public'
    """Access external and internal with READ and WRITE."""


class BaseFile:
    """Represents a single file on file system"""

    def __init__(self, path: str = None, contents_dirpath: str = None):
        self.file_path = str(path or __file__)
        self.file_ext = os.path.splitext(self.file_path)[1]

    @property
    def file_name(self):
        name, ext = os.path.splitext(os.path.basename(self.file_path))
        self.file_ext = ext
        return name

    @property
    def file_dirpath(self):
        return os.path.dirname(self.file_path)

    @property
    def file_dirname(self):
        return os.path.basename(os.path.dirname(self.file_path))

    @property
    def file_status(self) -> str:  # String
        return PageStatus.PROTECTED if self.file_name.startswith('_') else \
            PageStatus.FORBIDDEN if self.file_name.startswith('.') else PageStatus.PUBLIC

    @property
    def created_on(self):  # Datetime
        return datetime.fromtimestamp(os.path.getctime(self.file_path), tz=pytz.UTC)

    @property
    def modified_on(self):  # Datetime
        return datetime.fromtimestamp(os.path.getmtime(self.file_path), tz=pytz.UTC)

    def to_dict(self) -> dict[str, Any]:
        return self.__dict__


class BasePage:
    """Represents a single page returned from a web request"""
    _orm_options = {"mapper": {'created_on': 'file_created_on', 'modified_on': 'file_modified_on'}}
    # url: str = ''
    # slug: str = ''
    # title: str = ''
    # content: str = ''
    template: str = 'pages.html'
    created_on: datetime = None
    modified_on: datetime = None
    # entries: BasePagination = None

    def __lt__(self, other) -> bool:
        """Compares two BasePage instances based on their created_on attribute."""
        if not isinstance(other, BasePage):
            return True
        return self.created_on < other.created_on



import uuid
from datetime import datetime
from typing import Type, TypeVar, Dict, Iterable, Any

from sqlmodel import SQLModel
from sqlmodel.main import SQLModelMetaclass

from pyonir.utilities import deserialize_datestr

T = TypeVar("T")
class SchemaTable(SQLModelMetaclass):
    def __new__(cls, name, bases, namespace, **kwargs):
        # Grab frozen option (default False)
        is_frozen = kwargs.pop("frozen", None)
        is_frozen = is_frozen if is_frozen is not None else False

        # Build default model_config
        default_config = {
            "frozen": is_frozen,
            "from_attributes": False if is_frozen else True,
            "extra": "forbid" if is_frozen else "allow",
        }

        # Allow per-model override via namespace or kwargs
        user_config = namespace.pop("model_config", {})
        merged_config = {**default_config, **user_config}

        # Build the class
        new_cls = super().__new__(cls, name, bases, namespace, **kwargs)

        # Attach the merged config
        setattr(new_cls, "model_config", merged_config)

        return new_cls

class BaseSchema(SQLModel, metaclass=SchemaTable):
    """
    Interface for immutable dataclass models with CRUD and session support.
    """
    _errors: list[dict[str, Any]]

    def model_post_init(self, __context):
        object.__setattr__(self, "_errors", [])
        self.validate_fields()

    def __post_init__(self):
        self._errors = []
        self.validate_fields()


    def save_to_file(self, file_path: str) -> bool:
        """Saves the user data to a file in JSON format"""
        from pyonir.models.utils import create_file
        return create_file(file_path, self.to_dict(obfuscate=False))

    def save_to_session(self, request: 'PyonirRequest', key: str = None, value: any = None) -> None:
        """Convert instance to a serializable dict."""
        request.server_request.session[key or self.__class__.__name__.lower()] = value

    def to_dict(self, obfuscate = True):
        """Dictionary representing the instance"""

        obfuscated = lambda attr: obfuscate and hasattr(self,'_private_keys') and attr in (self._private_keys or [])
        is_ignored = lambda attr: attr.startswith("_") or callable(getattr(self, attr)) or obfuscated(attr)
        def process_value(key, value):
            if hasattr(value, 'to_dict'):
                return value.to_dict(obfuscate=obfuscate)
            if isinstance(value, property):
                return getattr(self, key)
            if isinstance(value, (tuple, list, set)):
                return [process_value(key, v) for v in value]
            return value
        fields = self.__class__.model_fields.keys() if hasattr(self.__class__, 'model_fields') else dir(self)
        return {key: process_value(key, getattr(self, key)) for key in fields if not is_ignored(key) and not obfuscated(key)}

    def to_json(self, obfuscate = True) -> str:
        """Returns a JSON serializable dictionary"""
        import json
        return json.dumps(self.to_dict(obfuscate))

    def is_valid(self) -> bool:
        """Returns True if there are no validation errors."""
        return not self._errors

    def validate_fields(self, field_name: str = None):
        """
        Validates fields by calling `validate_<fieldname>()` if defined.
        Clears previous errors on every call.
        """
        if field_name is not None:
            validator_fn = getattr(self, f"validate_{field_name}", None)
            if callable(validator_fn):
                validator_fn()
            return
        for name in self.__dict__.keys():
            if name.startswith("_"):
                continue
            validator_fn = getattr(self, f"validate_{name}", None)
            if callable(validator_fn):
                validator_fn()

    @classmethod
    def from_file(cls: Type[T], file_path: str, app_ctx=None) -> T:
        """Create an instance from a file path."""
        from pyonir.models.parser import DeserializeFile
        from pyonir.models.mapper import cls_mapper
        prsfile = DeserializeFile(file_path, app_ctx=app_ctx)
        return cls_mapper(prsfile, cls)

    @staticmethod
    def generate_date(date_value: str = None) -> datetime:
        return deserialize_datestr(date_value or datetime.now())

    @staticmethod
    def generate_id():
        return uuid.uuid4().hex


import os

from pyonir.models.database import BaseFSQuery, BasePagination
from pyonir.models.parser import DeserializeFile


test_files_dir = '/Users/hypermac/dev/pyonir/pyonir/tests/contents/pages'
query = BaseFSQuery(test_files_dir)

def test_init():
    assert query.sort_by == 'file_created_on'
    assert query.limit == 0
    assert query.max_count == 0
    assert query.curr_page == 0
    assert query.page_nums is None
    assert query.where_key is None
    assert query.sorted_files is None

def test_set_params():
    params = {
        'sort_by': 'file_name',
        'limit': '10',
        'curr_page': '1',
        'max_count': '100'
    }
    query.set_params(params)
    assert query.sort_by == 'file_name'
    assert query.limit == 10
    assert query.curr_page == 1
    assert query.max_count == 100

def test_paginated_collection():
    query.limit = 2
    query.curr_page = 1
    pagination = query.paginated_collection()

    assert isinstance(pagination, BasePagination)
    assert pagination.limit == 2
    assert pagination.curr_page == 1
    assert len(pagination.items) <= query.limit

def test_where_filter():
    # Test filtering by file name
    results = list(query.where('file_name', 'contains', 'index'))
    assert all('index' in file.file_name.lower() for file in results)

def test_prev_next():
    # Create a test file
    test_file = DeserializeFile(os.path.join(test_files_dir,"index.md"))
    result = BaseFSQuery.prev_next(test_file)

    assert hasattr(result, 'next')
    assert hasattr(result, 'prev')

def test_parse_params():
    # Test various parameter parsing cases
    assert BaseFSQuery.parse_params("name:value") == {"attr": "name", "op": "=", "value": "value"}
    assert BaseFSQuery.parse_params("age:>18") == {"attr": "age", "op": ">", "value": "18"}
    assert BaseFSQuery.parse_params("price:<=100") == {"attr": "price", "op": "<=", "value": "100"}



import pytest, os
from pyonir.models.parser import DeserializeFile
true = True
false = False
parselyFile = DeserializeFile(os.path.join(os.path.dirname(__file__),'contents', 'test.md'))


# test cases for DeserializeFile

def test_url():
    obj = "/"
    assert obj == parselyFile.data.get('url')


def test_slug():
    obj = ""
    assert obj == parselyFile.data.get('slug')


def test_inline_list_of_scalrs_types():
    obj = [1, true, "hello", 3.14, 1, true, "hello", 3.14]
    assert obj == parselyFile.data.get('inline_list_of_scalrs_types')


def test_single_item_list():
    obj = ["just one thing here"]
    assert obj == parselyFile.data.get('single_item_list')


def test_string_phonenumber():
    obj = "(111) 123-3456"
    assert obj == parselyFile.data.get('string_phonenumber')


def test_string_types():
    obj = "1, true, hello, 3.14"
    assert obj == parselyFile.data.get('string_types')


def test_basic():
    obj = "scalar value"
    assert obj == parselyFile.data.get('basic')


def test_dict_value():
    obj = {"my_key": "my_value", "another_key": "another_value"}
    assert obj == parselyFile.data.get('dict_value')


def test_list_value():
    obj = ["one", "two", "three"]
    assert obj == parselyFile.data.get('list_value')


def test_dynamic_list_blocks():
    obj = [{"ages": [1, true, "hello", 3.14, {"dict_key": "dict_value"}]}, {"this": {"age": 3, "key": "some value"}}]
    assert obj == parselyFile.data.get('dynamic_list_blocks')


def test_inline_list_of_maps():
    obj = [{"one": 1}, {"two": true}, {"three": "hello"}]
    assert obj == parselyFile.data.get('inline_list_of_maps')


def test_inline_dict_value():
    obj = "my_lnkey: my_lnvalue, another_lnkey: another_lnvalue"
    assert obj == parselyFile.data.get('inline_dict_value')


def test_multiline_block():
    obj = "What is this here? Content types enable you to organize and manage content in a consistent way for specific kinds of pages.\nthere is no such thing as a Python JSON object. JSON is a language independent file \nformat that finds its roots in JavaScript, and is supported by many languages. end of mulitiline block.\n"
    assert obj == parselyFile.data.get('multiline_block')


def test_js():
    obj = "if ('serviceWorker' in navigator) {\n  window.addEventListener('load', function() {\n    navigator.serviceWorker.register('/public/pwa/js/service-worker.js');\n  });\n}\n"
    assert obj == parselyFile.data.get('js')


def test_content():
    obj = "What is this here? Content types enable you to organize and manage content in a consistent way for specific kinds of pages.\nthere is no such thing as a Python JSON object. JSON is a language independent file \nformat that finds its roots in JavaScript, and is supported by many languages. If your YAML\n"
    assert obj == parselyFile.data.get('content')


def test_html():
    obj = "<app-screen>\n    <footer>\n        <span subtitle>Hello</span>\n        <img src=\"/public/some-image.jpg\" alt=\"find dibs logo\">\n        <button type=\"submit\">Join Pyonir</button>\n    </footer>\n</app-screen>\n"
    assert obj == parselyFile.data.get('html')




from pyonir.core import PyonirSchema

class EmailSubscriber(PyonirSchema, table="email_subscribers", private_keys=["email"]):
    """ Represents an email subscriber """

    def __init__(self, email: str, subscriptions: list[str] = None):
        self.email: str = email
        self.subscriptions: list[str] = subscriptions

    def validate_subscriptions(self):
        if not self.subscriptions:
            self._errors.append(f"Subscription is required")

    def validate_email(self):
        import re
        if not re.match(r"[^@]+@[^@]+\.[^@]+", self.email):
            self._errors.append(f"Invalid email address: {self.email}")


import os
from pyonir.models.user import User, UserMeta, Roles, UserSignIn

test_user_file = os.path.join(os.path.dirname(__file__), 'contents', 'test_user.json')
valid_credentials = {
    "email": "test@example.com",
    "password": "secure123"
}

def test_from_file():
    # Test loading user from file
    user = User.from_file(test_user_file)

    assert isinstance(user, User)
    assert user.email == "pyonir@site.com"
    assert user.name == "PyonirUserName"
    assert isinstance(user.meta, UserMeta)
    assert user.meta.first_name == "Test"
    assert user.meta.last_name == "User"
    assert user.role == "contributor"

def test_permissions_after_load():
    from pyonir.models.user import PermissionLevel
    user = User.from_file(test_user_file)

    # Test permissions based on role
    assert user.has_perm(PermissionLevel.READ)
    assert user.has_perm(PermissionLevel.WRITE)
    assert not user.has_perm(PermissionLevel.ADMIN)

def test_private_keys_excluded():
    user = User.from_file(test_user_file)
    serialized = user.to_dict()

    # Check private keys are excluded
    assert 'password' not in serialized
    assert 'auth_token' not in serialized
    assert 'id' not in serialized


# UserSignIn tests

def test_valid_signin():
    signin = UserSignIn(**valid_credentials)
    signin.validate_email()
    signin.validate_password()

    assert signin.is_valid()
    assert signin.email == valid_credentials["email"]
    assert signin.password == valid_credentials["password"]

def test_invalid_email_format():
    signin = UserSignIn(email="invalid-email", password="secure123")
    signin.validate_email()

    assert hasattr(signin, '_errors')
    assert "Invalid email address" in signin._errors[0]

def test_empty_email():
    signin = UserSignIn(email="", password="secure123")
    signin.validate_email()

    assert hasattr(signin, '_errors')
    assert not signin.is_valid()
    assert "Email cannot be empty" in signin._errors[0]
    test_invalid_email_format()

def test_empty_password():
    signin = UserSignIn(email="test@example.com", password="")
    signin.validate_password()

    assert hasattr(signin, '_errors')
    assert not signin.is_valid()
    assert "Password cannot be empty" in signin._errors[0]

def test_short_password():
    signin = UserSignIn(email="test@example.com", password="12345")
    signin.validate_password()

    assert hasattr(signin, '_errors')
    assert not signin.is_valid()
    assert "Password must be at least 6 characters long" in signin._errors[0]


from pyonir.pyonir_types import PyonirRoute, PyonirRouters
from .demo_controller import demo_items, subscriber_model, subscriber_values


# Define routes
routes: list[PyonirRoute] = [
    ('/items', demo_items, ["GET"]),
    ('/items/{sample_id:int}', demo_items, ["GET"]),
    ('/subscribe_values', subscriber_values, ["POST"]),
    ('/subscribe_model', subscriber_model, ["POST"]),
]

# Define an endpoint routers
router: PyonirRouters = [
    ('/api/demo', routes)
]


import os
import sqlite3
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, Optional, Type, Iterable, Callable, Union, Iterator, Generator

from sortedcontainers import SortedList

from pyonir.models.mapper import cls_mapper
from pyonir.models.parser import DeserializeFile
from pyonir.models.schemas import BaseSchema
from pyonir.pyonir_types import PyonirApp, AppCtx

@dataclass
class BasePagination:
    limit: int = 0
    max_count: int = 0
    curr_page: int = 0
    page_nums: list[int, int] = field(default_factory=list)
    items: list['DeserializeFile'] = field(default_factory=list)

    def __iter__(self) -> Iterator['DeserializeFile']:
        return iter(self.items)

class DatabaseService(ABC):
    """Stub implementation of DatabaseService with env-based config + builder overrides."""

    def __init__(self, app: PyonirApp) -> None:
        # Base config from environment
        from pyonir.utilities import get_attr
        self._config: object = get_attr(app.env, 'database')
        self.connection: Optional[sqlite3.Connection] = None
        self._database: str = '' # the db address or name. path/to/directory, path/to/sqlite.db
        self._driver: str = '' #the db context fs, sqlite, mysql, pgresql, oracle
        self._host: str = ''
        self._port: int = 0
        self._username: str = ''
        self._password: str = ''

    @property
    def driver(self) -> Optional[str]:
        return self._driver

    @property
    def host(self) -> Optional[str]:
        return self._host

    @property
    def port(self) -> Optional[int]:
        return self._port

    @property
    def username(self) -> Optional[str]:
        return self._username

    @property
    def password(self) -> Optional[str]:
        return self._password

    @property
    def database(self) -> Optional[str]:
        return self._database

    # --- Builder pattern overrides ---
    def set_driver(self, driver: str) -> "DatabaseService":
        self._driver = driver
        return self

    def set_database(self, database: str) -> "DatabaseService":
        self._database = database
        return self

    def set_host(self, host: str) -> "DatabaseService":
        self._host = host
        return self

    def set_port(self, port: int) -> "DatabaseService":
        self._port = port
        return self

    def set_username(self, username: str) -> "DatabaseService":
        self._username = username
        return self

    def set_password(self, password: str) -> "DatabaseService":
        self._password = password
        return self

    # --- Database operations ---
    @abstractmethod
    def connect(self) -> None:
        if not self.database:
            raise ValueError("Database must be set before connecting")

        if self.driver == "sqlite":
            print(f"[DEBUG] Connecting to SQLite database at {self.database}")
            self.connection = sqlite3.connect(self.database)
            self.connection.row_factory = sqlite3.Row
        elif self.driver == "fs":
            print(f"[DEBUG] Using file system path at {self.database}")
            Path(self.database).mkdir(parents=True, exist_ok=True)
        else:
            raise ValueError(f"Unknown driver: {self.driver}")

    @abstractmethod
    def disconnect(self) -> None:
        print(f"[DEBUG] Disconnecting from {self.driver}:{self.database}")
        if self.driver == "sqlite" and self.connection:
            self.connection.close()
            self.connection = None
        # FS needs to reset its database location

    # @abstractmethod
    # def execute_query(self, query: str, params: Optional[Dict[str, Any]] = None) -> Any:
    #     print(f"[DEBUG] Executing query: {query} with params: {params}")
    #     return NotImplemented

    @abstractmethod
    def insert(self, table: str, entity: Type[BaseSchema]) -> Any:
        """Insert entity into backend."""
        table = table or entity.__class__.__name__.lower()
        data = entity.to_dict()

        if self.driver == "sqlite":
            keys = ', '.join(data.keys())
            placeholders = ', '.join('?' for _ in data)
            query = f"INSERT INTO {table} ({keys}) VALUES ({placeholders})"
            cursor = self.connection.cursor()
            cursor.execute(query, tuple(data.values()))
            self.connection.commit()
            return cursor.lastrowid

        elif self.driver == "fs":
            # Save JSON file per record
            entity.save_to_file(entity.file_path)
            return os.path.exists(entity.file_path)

    @abstractmethod
    def find(self, entity_cls: Type[BaseSchema], filter: Dict = None) -> Any:
        import json
        table = entity_cls.__name__.lower()
        results = []

        if self.driver == "sqlite":
            where_clause = ''
            params = ()
            if filter:
                where_clause = 'WHERE ' + ' AND '.join(f"{k} = ?" for k in filter)
                params = tuple(filter.values())
            query = f"SELECT * FROM {table} {where_clause}"
            cursor = self.connection.cursor()
            cursor.execute(query, params)
            results = [dict(row) for row in cursor.fetchall()]

        elif self.driver == "fs":
            table_path = Path(self.database) / table
            if table_path.exists():
                for file_path in table_path.glob("*.json"):
                    with open(file_path) as f:
                        record = json.load(f)
                        if filter:
                            match = all(record.get(k) == v for k, v in filter.items())
                            if match:
                                results.append(record)
                        else:
                            results.append(record)

        return results



class BaseFSQuery:
    """Base class for querying files and directories"""
    _cache: Dict[str, Any] = {}

    def __init__(self, query_path: str,
                app_ctx: AppCtx = None,
                model: Optional[object] = None,
                name_pattern: str = None,
                exclude_dirs: tuple = None,
                exclude_names: tuple = None,
                force_all: bool = True) -> None:

        self.query_path = query_path
        self.sort_by: str = 'file_created_on'
        self.limit: int = 0
        self.max_count: int = 0
        self.curr_page: int = 0
        self.page_nums: list[int, int] = None
        self.where_key: str = None
        self.sorted_files: SortedList = None
        self.files: Generator[DeserializeFile] = query_fs(query_path,
                              app_ctx = app_ctx,
                              model = model,
                              name_pattern = name_pattern,
                              exclude_dirs = exclude_dirs,
                              exclude_names = exclude_names,
                              force_all = force_all)

    def set_params(self, params: dict):
        for k in ['sort_by', 'limit', 'curr_page','max_count','page_nums']:
            if k in params:
                if k in ('limit', 'curr_page', 'max_count') and params[k]:
                    params[k] = int(params[k])
                setattr(self, k, params[k])
        return self

    def paginated_collection(self)-> Optional[BasePagination]:
        """Paginates a list into smaller segments based on curr_pg and display limit"""
        from sortedcontainers import SortedList
        from pyonir.models.utils import get_attr

        if self.sort_by:
            self.sorted_files = SortedList(self.files, lambda x: get_attr(x, self.sort_by) or x)
        if self.where_key:
            where_key = [self.parse_params(ex) for ex in self.where_key.split(',')]
            self.sorted_files = SortedList(self.where(**where_key[0]), lambda x: get_attr(x, self.sort_by))
        force_all = not self.limit

        self.max_count = len(self.sorted_files)
        page_num = 0 if force_all else int(self.curr_page)
        start = (page_num * self.limit) - self.limit
        end = (self.limit * page_num)
        pg = (self.max_count // self.limit) + (self.max_count % self.limit > 0) if self.limit > 0 else 0
        pag_data = self.paginate(start=start, end=end, reverse=True) if not force_all else self.sorted_files

        return BasePagination(
            curr_page = page_num,
            page_nums = [n for n in range(1, pg + 1)] if pg else None,
            limit = self.limit,
            max_count = self.max_count,
            items = list(pag_data)
        )

    def paginate(self, start: int, end: int, reverse: bool = False):
        """Returns a slice of the items list"""
        sl = self.sorted_files.islice(start, end, reverse=reverse) if end else self.sorted_files
        return sl

    @staticmethod
    def prev_next(input_file: 'DeserializeFile'):
        """Returns the previous and next files relative to the input file"""
        from pyonir.models.mapper import dict_to_class
        prv = None
        nxt = None
        pc = BaseFSQuery(input_file.file_dirpath)
        _collection = iter(pc.files)
        for cfile in _collection:
            if cfile.file_status == 'hidden': continue
            if cfile.file_path == input_file.file_path:
                nxt = next(_collection, None)
                break
            else:
                prv = cfile
        return dict_to_class({"next": nxt, "prev": prv})

    def find(self, value: any, from_attr: str = 'file_name'):
        """Returns the first item where attr == value"""
        return next((item for item in self.sorted_files if getattr(item, from_attr, None) == value), None)

    def where(self, attr, op="=", value=None):
        """Returns a list of items where attr == value"""
        from pyonir.models.utils import get_attr
        # if value is None:
        #     # assume 'op' is actually the value if only two args were passed
        #     value = op
        #     op = "="

        def match(item):
            actual = get_attr(item, attr)
            if not hasattr(item, attr):
                return False
            if actual and not value:
                return True # checking only if item has an attribute
            elif op == "=":
                return actual == value
            elif op == "in" or op == "contains":
                return actual in value if actual is not None else False
            elif op == ">":
                return actual > value
            elif op == "<":
                return actual < value
            elif op == ">=":
                return actual >= value
            elif op == "<=":
                return actual <= value
            elif op == "!=":
                return actual != value
            return False
        if callable(attr): match = attr
        if not self.sorted_files:
            self.sorted_files = SortedList(self.files, lambda x: get_attr(x, self.sort_by) or x)
        return filter(match, list(self.sorted_files or self.files))

    def __len__(self):
        return self.sorted_files and len(self.sorted_files) or 0

    def __iter__(self):
        return iter(self.sorted_files)

    @staticmethod
    def parse_params(param: str):
        k, _, v = param.partition(':')
        op = '='
        is_eq = lambda x: x[1]=='='
        if v.startswith('>'):
            eqs = is_eq(v)
            op = '>=' if eqs else '>'
            v = v[1:] if not eqs else v[2:]
        elif v.startswith('<'):
            eqs = is_eq(v)
            op = '<=' if eqs else '<'
            v = v[1:] if not eqs else v[2:]
            pass
        else:
            pass
        return {"attr": k.strip(), "op":op, "value":BaseFSQuery.coerce_bool(v)}

    @staticmethod
    def coerce_bool(value: str):
        d = ['false', 'true']
        try:
            i = d.index(value.lower().strip())
            return True if i else False
        except ValueError as e:
            return value.strip()


def query_fs(abs_dirpath: str,
                app_ctx: AppCtx = None,
                model: Union[object, str] = None,
                name_pattern: str = None,
                exclude_dirs: tuple = None,
                exclude_names: tuple = None,
                force_all: bool = True) -> Generator:
    """Returns a generator of files from a directory path"""
    from pathlib import Path
    from pyonir.models.page import BasePage, DeserializeFile
    from pyonir.models.media import BaseMedia

    # results = []
    hidden_file_prefixes = ('.', '_', '<', '>', '(', ')', '$', '!', '._')
    allowed_content_extensions = ('prs', 'md', 'json', 'yaml')
    def get_datatype(filepath) -> Union[object, BasePage, BaseMedia]:
        if model == 'path': return str(filepath)
        if model == BaseMedia: return BaseMedia(filepath)
        pf = DeserializeFile(str(filepath), app_ctx=app_ctx)
        if model == 'file':
            return pf
        pf.schema = BasePage if (pf.is_page and not model) else model
        res = cls_mapper(pf, pf.schema) if pf.schema else pf
        return res

    def skip_file(file_path: Path) -> bool:
        """Checks if the file should be skipped based on exclude_dirs and exclude_file"""
        is_private_file = file_path.name.startswith(hidden_file_prefixes)
        is_excluded_file = exclude_names and file_path.name in exclude_names
        is_allowed_file = file_path.suffix[1:] in allowed_content_extensions
        if not is_private_file and force_all: return False
        return is_excluded_file or is_private_file or not is_allowed_file

    for path in Path(abs_dirpath).rglob(name_pattern or "*"):
        if skip_file(path) or path.is_dir(): continue
        yield get_datatype(path)